from __future__ import division, unicode_literals, absolute_import

from builtins import int, range, dict, map, zip

import os
import io
import re
import sys
import mappy
import queue
import threading

# Future warning from cython in h5py
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import h5py

import numpy as np
np.seterr(all='raise')

from time import sleep
from operator import itemgetter
from collections import defaultdict
from multiprocessing import Process, Pipe

if sys.version_info[0] > 2:
    unicode = str

# import tombo modules/functions
from . import tombo_stats as ts
from . import tombo_helper as th

from ._default_parameters import (
    SEG_PARAMS_TABLE, ALGN_PARAMS_TABLE, EXTRA_SIG_FACTOR, MASK_FILL_Z_SCORE,
    MASK_BASES, START_BANDWIDTH, START_SEQ_WINDOW, BAND_BOUNDARY_THRESH,
    DEL_FIX_WINDOW, MAX_DEL_FIX_WINDOW, MIN_EVENT_TO_SEQ_RATIO, MAX_RAW_CPTS)

from .dynamic_programming import traceback, forward_pass
from .c_helper import (
    c_new_means, c_valid_cpts_w_cap, c_valid_cpts_w_cap_t_test)
from .c_dynamic_programming import (
    c_reg_z_scores, c_banded_forward_pass, c_banded_traceback,
    c_base_z_scores, c_adaptive_banded_forward_pass)

VERBOSE = False
PROGRESS_INTERVAL = 1000
PROC_UPDATE_INTERVAL = max(1, int(PROGRESS_INTERVAL / 100))

_PROFILE_RSQGL = False

_DEBUG_FIT = False
_DEBUG_FULL = False
_DEBUG_MIDDLE = False
_DEBUG_PARAMS = False
_DRY_RUN = any((_DEBUG_PARAMS, _DEBUG_FIT, _DEBUG_FULL, _DEBUG_MIDDLE))
_NUM_DEBUG_ENDS = 250


###############################################
########## Read Segmentation Scoring ##########
###############################################

def get_read_seg_score(norm_signal, segs, r_ref_means, r_ref_sds):
    return np.mean([
        np.abs((b_m - b_ref_m) / b_ref_s)
        for b_m, b_ref_m, b_ref_s in
        zip(c_new_means(norm_signal, segs), r_ref_means, r_ref_sds)])


##################################
########## Debug Output ##########
##################################

def _write_middle_debug(z_scores, fwd_pass, band_event_starts,
                       debug_fp, reg_id, debug_num_seq=_NUM_DEBUG_ENDS,
                       short=False):
    fwd_pass = fwd_pass[1:]
    if fwd_pass.shape[0] < debug_num_seq:
        debug_num_seq = fwd_pass.shape[0]
    debug_end_start = len(band_event_starts) - debug_num_seq
    debug_fp.write('\n'.join(
        '\t'.join(map(str, (band_pos + band_event_starts[seq_pos], seq_pos,
                            score, unicode(reg_id) + 'z_begin')))
        for seq_pos, s_data in enumerate(z_scores[:debug_num_seq])
        for band_pos, score in enumerate(s_data)) + '\n')
    debug_fp.write('\n'.join(
        '\t'.join(map(str, (band_pos + band_event_starts[seq_pos], seq_pos,
                            score, unicode(reg_id) + 'fwd_begin')))
        for seq_pos, s_data in enumerate(fwd_pass[:debug_num_seq])
        for band_pos, score in enumerate(s_data)) + '\n')
    if short: return

    debug_fp.write('\n'.join(
        '\t'.join(map(str, (
            band_pos + band_event_starts[debug_end_start + seq_pos], seq_pos,
            score, unicode(reg_id) + 'z_end')))
        for seq_pos, s_data in enumerate(z_scores[-debug_num_seq:])
        for band_pos, score in enumerate(s_data)) + '\n')
    debug_fp.write('\n'.join(
        '\t'.join(map(str, (
            band_pos + band_event_starts[debug_end_start + seq_pos], seq_pos,
            score, unicode(reg_id) + 'fwd_end')))
        for seq_pos, s_data in enumerate(fwd_pass[-debug_num_seq:])
        for band_pos, score in enumerate(s_data)) + '\n')

    return

def _write_full_debug(fwd_pass_move, band_event_starts, top_max_pos,
                     z_scores, debug_fp, failed_fp, reg_id, final_score):
    read_tb = c_banded_traceback(fwd_pass_move, band_event_starts, top_max_pos)
    prev_event_pos = read_tb[0]
    band_poss = []
    event_scores = []
    for seq_pos, event_pos in enumerate(read_tb[1:]):
        seq_band_poss = [e_pos - band_event_starts[seq_pos]
                         for e_pos in range(prev_event_pos, event_pos)]
        band_poss.extend(seq_band_poss)
        event_scores.extend([z_scores[seq_pos][b_pos]
                             for b_pos in seq_band_poss])
        prev_event_pos = event_pos

    debug_fp.write('\n'.join(
        '\t'.join(map(str, (e_pos, b_pos, e_score, unicode(reg_id))))
        for e_pos, (b_pos, e_score) in enumerate(zip(
                band_poss, event_scores))) + '\n')
    fail_str = (('Failed ' if final_score < 0 else 'Pass ') +
                unicode(final_score) + ' ' + unicode(final_score / len(read_tb)))
    failed_fp.write(fail_str + '\t' + unicode(reg_id) + '\n')

    return

def _write_tb_debug(fwd_pass_move, band_event_starts, top_max_pos,
                   debug_fp, reg_id, debug_num_seq=_NUM_DEBUG_ENDS):
    read_tb = c_banded_traceback(fwd_pass_move, band_event_starts, top_max_pos)
    debug_fp.write('\n'.join(
        '\t'.join(map(str, (event_pos, seq_pos,
                            unicode(reg_id) + 'fwd_begin')))
        for seq_pos, event_pos in enumerate(read_tb[:debug_num_seq])) + '\n')
    debug_fp.write('\n'.join(
        '\t'.join(map(str, (event_pos, seq_pos,
                            unicode(reg_id) + 'fwd_end')))
        for seq_pos, event_pos in enumerate(read_tb[-debug_num_seq:])) + '\n')
    return

def _write_fit_debug(
        norm_signal, segs, r_ref_means, r_ref_sds, genome_seq):
    norm_means = c_new_means(norm_signal, segs)
    with io.open('eventless_testing.model.txt', 'wt') as fp:
        fp.write('Position\tMean\tSD\n' + '\n'.join(
            '\t'.join(map(str, (pos, p_mean, p_std)))
            for pos, (p_mean, p_std) in enumerate(zip(
                    r_ref_means, r_ref_sds))) + '\n')
    with io.open('eventless_testing.seq.txt', 'wt') as fp:
        fp.write('Base\tPosition\tSignalMean\n' + '\n'.join(
            '\t'.join(map(str, (b, pos, p_mean))) for pos, (b, p_mean) in
            enumerate(zip(genome_seq, norm_means))) + '\n')
    Position, Signal = [], []
    for base_i, (b_start, b_end) in enumerate(zip(segs[:-1], segs[1:])):
        Position.extend(
            base_i + np.linspace(0, 1, b_end - b_start, endpoint=False))
        Signal.extend(norm_signal[b_start:b_end])
    with io.open('eventless_testing.signal.txt', 'wt') as fp:
        fp.write('Position\tSignal\n' + '\n'.join(
            '\t'.join(map(str, (pos, sig)))
            for pos, sig in zip(Position, Signal)) + '\n')

    return

def _write_params_debug(
        norm_signal, segs, r_ref_means, r_ref_sds,
        running_stat_width, min_obs_per_base, mean_obs_per_event,
        match_evalue, skip_pen, bandwidth, fast5_fn):
    mean_half_z_score = get_read_seg_score(
        norm_signal, segs, r_ref_means, r_ref_sds)
    sys.stdout.write(
        '\t'.join(map(str, (
            running_stat_width, min_obs_per_base, mean_obs_per_event,
            match_evalue, skip_pen, bandwidth, fast5_fn,
            mean_half_z_score))) + '\n')

    return

def _open_debug_fps():
    score_fp = io.open('debug_event_align.txt', 'wt')
    score_fp.write('EventPos\tSeqPos\tScore\tRegion\n')
    tb_fp = io.open('debug_event_align.traceback.txt', 'wt')
    tb_fp.write('EventPos\tSeqPos\tRegion\n')
    full_fit_fp = io.open('debug_event_align.full_fit.txt', 'wt')
    full_fit_fp.write('EventPos\tBandPos\tEventScore\tRegion\n')
    full_failed_fp = io.open('debug_event_align.full_failed.txt', 'wt')
    full_failed_fp.write('DidFail\tRegion\n')
    debug_fps = [score_fp, tb_fp, full_fit_fp, full_failed_fp]

    return debug_fps


############################################
########## Raw Signal Re-squiggle ##########
############################################

def get_model_fit_segs(
        segs, norm_signal, r_ref_means, r_ref_sds, min_obs_per_base,
        max_raw_cpts=None, del_fix_window=DEL_FIX_WINDOW,
        max_del_fix_window=MAX_DEL_FIX_WINDOW,
        extra_sig_factor=EXTRA_SIG_FACTOR):
    """
    Find new segments at skipped bases during dynamic programming segmentation.

    :param segs: current Read segment locations
    :param norm_signal: Normalized read siganl
    :param r_ref_means: Read refererence means from genomic sequence
    :param r_ref_sds: Read refererence standard deviations from genomic sequence
    :param min_obs_per_base: Minimum raw observations to assign to each base
    :param max_raw_cpts: Maximum new changepoints to find from raw signal
    :param del_fix_window: initial bases to extend skipped base windows
    :param max_del_fix_window: max bases to extend skipped base windows
    :param extra_sig_factor: Amount of extra signal to require in order to
        perform signal space re-squiggle

    :returns: New segments with skipped bases resolved
    """
    def merge_del_windows(all_del_windows):
        merged_del_windows = []
        for start, end in all_del_windows:
            if (len(merged_del_windows) > 0 and
                start < merged_del_windows[-1][1]):
                merged_del_windows[-1] = (merged_del_windows[-1][0], end)
            else:
                merged_del_windows.append((start, end))

        return merged_del_windows

    def window_too_small(start, end):
        n_events = end - start
        sig_start, sig_end = segs[start], segs[end]
        sig_len = sig_end - sig_start
        # windows are expanded by one base and the extra signal factor
        # to allow some room to search for best path
        return sig_len <= ((n_events + 1) * min_obs_per_base) * extra_sig_factor

    def expand_small_windows(all_del_windows):
        expanded_del_windows = []
        windows_expanded = False
        for start, end in all_del_windows:
            if window_too_small(start, end):
                windows_expanded = True
                start -= 1
                end += 1
            expanded_del_windows.append((start, end))

        return expanded_del_windows, windows_expanded

    def trim_del_window_ends(all_del_windows):
        # potentially trim first and last windows
        if all_del_windows[0][0] < 0:
            all_del_windows[0] = (0, all_del_windows[0][1])
        if all_del_windows[-1][1] > len(segs) - 1:
            all_del_windows[-1] = (all_del_windows[-1][0], len(segs) - 1)

        return all_del_windows

    def get_deletion_windows():
        # get initial windows around deletions/skipped bases
        all_del_windows = []
        for del_pos in np.where(np.diff(segs) == 0)[0]:
            if (len(all_del_windows) > 0 and
                del_pos < all_del_windows[-1][1] + del_fix_window):
                all_del_windows[-1] = (all_del_windows[-1][0],
                                       del_pos + del_fix_window + 1)
            else:
                all_del_windows.append((del_pos - del_fix_window,
                                        del_pos + del_fix_window + 1))
        if len(all_del_windows) == 0:
            return

        windows_expanded = False
        all_del_windows = merge_del_windows(all_del_windows)
        all_del_windows = trim_del_window_ends(all_del_windows)
        # expand small windows until there are no more or the max del window
        # expansions have been attempted.
        for _ in range(max_del_fix_window - del_fix_window):
            all_del_windows, windows_expanded = expand_small_windows(
                all_del_windows)
            if not windows_expanded: break
            all_del_windows = merge_del_windows(all_del_windows)
            all_del_windows = trim_del_window_ends(all_del_windows)

        if windows_expanded and any(
                window_too_small(start, end) for start, end in all_del_windows):
            raise NotImplementedError(
                'Not enough raw signal around potential genomic deletion(s)')

        if max_raw_cpts is not None and max([
                end - start for start, end in all_del_windows]) > max_raw_cpts:
            raise NotImplementedError(
                'Read contains too many potential genomic deletions')

        return all_del_windows


    all_del_windows = get_deletion_windows()
    if all_del_windows is None:
        return segs

    for start, end in all_del_windows:
        n_events = end - start
        sig_start, sig_end = segs[start], segs[end]
        sig_len = sig_end - sig_start

        # find signal space z-scores mapping without real banding by allowing
        # entire window to be searched (c_reg_z_scores will clip base search
        # windows to enforce min_obs_per_base)
        pseudo_starts = np.linspace(0, sig_len, n_events + 1, dtype=np.int64)
        reg_z_scores = c_reg_z_scores(
            norm_signal[sig_start:sig_end], r_ref_means[start:end],
            r_ref_sds[start:end], pseudo_starts,
            0, n_events, n_events, min_obs_per_base)
        reg_fwd_scores = forward_pass(reg_z_scores, min_obs_per_base)
        # perform signal based scoring segmentation
        #  - it is ~60X faster than base space
        reg_segs = traceback(reg_fwd_scores, min_obs_per_base) + sig_start
        assert reg_segs.shape[0] == end - start - 1
        segs[start+1:end] = reg_segs

    if np.diff(segs).min() < 1:
        raise NotImplementedError('New segments include zero length events')
    if segs[0] < 0:
        raise NotImplementedError('New segments start with negative index')
    if segs[-1] > norm_signal.shape[0]:
        raise NotImplementedError('New segments end past raw signal values')

    return segs


#####################################################
########## Static Band Dynamic Programming ##########
#####################################################

def get_short_read_event_mapping(
        event_means, r_ref_means, r_ref_sds, skip_pen, stay_pen, z_shift,
        reg_id=None, debug_fps=None):
    """
    Perform banded dynamic programming sequence to event alignment
    without masking

    :param event_means: Numpy array with read base means
    :param r_ref_means: Numpy array with read reference means
    :param r_ref_sds: Numpy array with read reference standard deviations
    :param skip_pen: Penalty applied to skipped genomic bases
    :param stay_pen: Penalty applied to stay states (should shift to 0
        expected value)
    :param z_shift: Shift z-scores by this amount (includes matching
        positive expected value)

    :returns: Event to sequence mapping for full length of short read
    """
    seq_len = r_ref_means.shape[0]
    events_len = event_means.shape[0]

    # create read starts in order to clip just the corners of the full events to
    # seqeunce matrix
    mask_len = min(seq_len, events_len) // 4
    band_event_starts = np.concatenate([
        np.zeros(seq_len - mask_len * 2),
        np.linspace(0, mask_len, mask_len * 2)]).astype(np.int64)
    bandwidth = events_len - mask_len

    shifted_z_scores = z_shift - np.row_stack([
        np.abs(event_means[event_pos:event_pos + bandwidth]
               - r_ref_means[seq_pos]) / r_ref_sds[seq_pos]
        for seq_pos, event_pos in enumerate(band_event_starts)])
    fwd_pass, fwd_pass_move = c_banded_forward_pass(
        shifted_z_scores, band_event_starts, skip_pen, stay_pen)

    # perform traceback
    top_max_pos = np.argmax(fwd_pass[-1,:])
    if _DEBUG_FULL:
        _write_full_debug(fwd_pass_move, band_event_starts, top_max_pos,
                         shifted_z_scores, debug_fps[2], debug_fps[3], reg_id,
                         fwd_pass[-1,top_max_pos])
    if _DEBUG_MIDDLE:
        _write_middle_debug(shifted_z_scores, fwd_pass, band_event_starts,
                           debug_fps[0], reg_id, short=True)
        _write_tb_debug(fwd_pass_move, band_event_starts, top_max_pos,
                       debug_fps[1], reg_id)

    read_tb = c_banded_traceback(fwd_pass_move, band_event_starts, top_max_pos)

    return read_tb


#######################################################
########## Adaptive Band Dynamic Programming ##########
#######################################################

def get_masked_start_fwd_pass(
        event_means, r_ref_means, r_ref_sds, mapped_start_offset,
        skip_pen, stay_pen, z_shift, bandwidth, events_per_base,
        mask_fill_z_score=MASK_FILL_Z_SCORE,
        mask_bases=MASK_BASES, reg_id=None, debug_fps=None):
    """
    Perform banded dynamic programming sequence to event alignment forcing
    the path to start and end at the previously discovered locations.
    This is performed by masking the z-scores outside a "cone" extended
    mask_bases from the beginning and end of the middle of the read.

    :param event_means: Numpy array with read base means
    :param r_ref_means: Numpy array with read reference means
    :param r_ref_sds: Numpy array with read reference standard deviations
    :param mapped_start_offset: Previously identified start of genomic
        sequence within events
    :param skip_pen: Penalty applied to skipped genomic bases
    :param stay_pen: Penalty applied to stay states (should shift to 0
        expected value)
    :param z_shift: Shift z-scores by this amount (includes matching positive
        expected value)
    :param bandwidth: Bandwidth over which to search for sequence to
        event mapping
    :param events_per_base: Average events per base for the start mapping

    :returns: Event to sequence mapping for start of read including forward
        pass scores, forward pass move
        values, band starts within the events vector and z-scores
    """
    assert event_means.shape[0] - mapped_start_offset >= bandwidth, (
        'Read sequence to signal matching starts too far into events for ' +
        'full adaptive assignment.')
    half_bandwidth = bandwidth // 2

    # check if the mapped start position is too close to the end of
    # the events array and extend the bandwidth window if so
    band_events_start_pos = (
        0 if half_bandwidth <= mapped_start_offset else
        mapped_start_offset - half_bandwidth)

    tmp_seq_len = max(half_bandwidth, mask_bases,
                      int((half_bandwidth + 1) / events_per_base)) + 1
    band_event_starts = np.linspace(
        band_events_start_pos,
        band_events_start_pos + (tmp_seq_len  * events_per_base),
        tmp_seq_len).astype(np.int64)
    mask_seq_len = max(
        mask_bases, next(i + 2 for i, bes in enumerate(band_event_starts)
                         if bes >= mapped_start_offset))
    band_event_starts = band_event_starts[:mask_seq_len]

    # get masked z-scores at the beginning of the read
    mask_start_pos = np.linspace(
        mapped_start_offset + 1,
        band_event_starts[mask_bases - 1] + bandwidth,
        mask_bases).astype(np.int64)
    def get_start_mask_z_score(seq_pos, event_pos):
        start_mask_len = max(mapped_start_offset - event_pos, 0)
        end_mask_len = (0 if seq_pos >= mask_bases else
                        bandwidth - (mask_start_pos[seq_pos] - event_pos))
        event_vals = event_means[event_pos + start_mask_len:
                                 event_pos + bandwidth - end_mask_len]
        b_z_scores = c_base_z_scores(
            event_vals, r_ref_means[seq_pos], r_ref_sds[seq_pos])
        masked_z_scores = np.concatenate([
            [mask_fill_z_score] * start_mask_len, b_z_scores,
            [mask_fill_z_score] * end_mask_len])
        del b_z_scores
        return masked_z_scores
    shifted_z_scores = np.empty((band_event_starts.shape[0], bandwidth))
    for seq_pos, event_pos in enumerate(band_event_starts):
        shifted_z_scores[seq_pos,:] = get_start_mask_z_score(seq_pos, event_pos)
    shifted_z_scores += z_shift
    fwd_pass, fwd_pass_move = c_banded_forward_pass(
        shifted_z_scores, band_event_starts, skip_pen, stay_pen)

    return fwd_pass, fwd_pass_move, band_event_starts, shifted_z_scores

def get_mapping_start(
        event_means, r_ref_means, r_ref_sds, skip_pen, stay_pen, z_shift,
        seq_window, bandwidth, norm_signal, valid_cpts, score_thresh,
        min_obs_per_base, reg_id=None, debug_fps=None):
    """
    Perform banded dynamic programming sequence to event alignment through
    The beginning of an read to identify the start of genome sequence to
    event matching

    :param event_means: Numpy array with read base means
    :param r_ref_means: Numpy array with read reference means
    :param r_ref_sds: Numpy array with read reference standard deviations
    :param skip_pen: Penalty applied to skipped genomic bases
    :param stay_pen: Penalty applied to stay states (should shift to 0
        expected value)
    :param z_shift: Shift z-scores by this amount (includes matching positive
        expected value)
    :param seq_window: Number of genomic bases to search over for the start of
        the read
    :param bandwidth: Bandwidth over which to search for sequence to
        event mapping
    :param norm_signal: Normalized raw signal vector
    :param valid_cpts: Segmentation positions within norm_signal
    :param score_thresh: Read mean half-normal signal segmentation score
        threshold

    :returns: Start position (0-based) of seqeunce to event alignment within
        events and the mean events_per_base through the queried portion of read
    """
    if event_means.shape[0] < bandwidth:
        raise NotImplementedError('Read too short for start/end discovery')
    if r_ref_means.shape[0] < seq_window:
        raise NotImplementedError(
            'Genomic mapping too short for start/end discovery')

    # banded z-scores (moving up one event per base for start/end discovery
    start_z_scores = z_shift - np.row_stack([
        np.abs(event_means[seq_pos:seq_pos + bandwidth] -
               r_ref_means[seq_pos]) / r_ref_sds[seq_pos]
        for seq_pos in range(seq_window)])
    start_band_event_starts = np.linspace(
        0, seq_window, seq_window).astype(np.int64)

    np.arange(seq_window, dtype=np.int64)

    start_fwd_pass, start_fwd_pass_move = c_banded_forward_pass(
        start_z_scores, start_band_event_starts, skip_pen, stay_pen)

    # find max along the top and right edges to start traceback
    top_max_pos = np.argmax(start_fwd_pass[-1,:])

    # perform traceback
    start_tb = c_banded_traceback(
        start_fwd_pass_move, start_band_event_starts, top_max_pos)

    # check that read start mapping is valid to avoid wasted compute on
    # adaptive dp
    start_segs = valid_cpts[start_tb]
    start_sig = norm_signal[start_segs[0]:start_segs[-1]]
    start_segs = start_segs - start_segs[0]
    start_segs = get_model_fit_segs(
        start_segs, start_sig, r_ref_means[:seq_window],
        r_ref_sds[:seq_window], min_obs_per_base)
    if get_read_seg_score(
            start_sig, start_segs, r_ref_means[:seq_window],
            r_ref_sds[:seq_window]) > score_thresh:
        raise NotImplementedError(
            'Poor raw to expected signal matching at read start')

    # compute the average events per base to use for the start forward pass
    events_per_base = (start_tb[-1] - start_tb[0]) / len(start_tb)
    start_loc = start_tb[0]

    return start_loc, events_per_base

def find_adaptive_base_assignment(
        norm_signal, running_stat_width, min_obs_per_base, num_events, std_ref,
        genome_seq, genome_loc, skip_pen, stay_pen, z_shift, bandwidth, is_rna,
        score_thresh, start_bandwidth=START_BANDWIDTH,
        start_seq_window=START_SEQ_WINDOW,
        band_boundary_thresh=BAND_BOUNDARY_THRESH, reg_id=None, debug_fps=None):
    """
    Perform banded dynamic programming sequence to event alignment by first
    identifying the start of the sequence to event matching and then
    performing banded matching through the whole read

    :param norm_signal: Numpy array with normalized read signal
    :param running_stat_width: Width of neighboring windows over which to
        compute changepoint stats
    :param min_obs_per_base: Minimum number of raw observations per base
    :param num_events: Number of events to identify in this read
    :param std_ref: A TomboModel object
    :param genome_seq: Genomic sequence for this read
    :param genome_loc: Mapped genomic location for this read
    :param skip_pen: Penalty applied to skipped genomic bases
    :param stay_pen: Penalty applied to stay states (should shift to 0
        expected value)
    :param z_shift: Shift z-scores by this amount (includes matching positive
        expected value)
    :param bandwidth: Bandwidth over which to search for sequence to
        event mapping
    :param is_rna: Is this an RNA read
    :param score_thresh: Read mean half-normal segmentation score threshold

    :returns: Start of seqeunce to event alignment and the mean
        events_per_base through the queried portion of a read
    """
    # get events
    if is_rna:
        # RNA bases show consistent variable spread so use t-test segmentation
        valid_cpts = c_valid_cpts_w_cap_t_test(
            norm_signal, min_obs_per_base, running_stat_width, num_events)
    else:
        valid_cpts = c_valid_cpts_w_cap(
            norm_signal, min_obs_per_base, running_stat_width, num_events)
        #valid_cpts = th.get_valid_cpts(
        #    norm_signal, running_stat_width, num_events)
    valid_cpts.sort()
    event_means = c_new_means(norm_signal, valid_cpts)

    dnstrm_bases = std_ref.kmer_width - std_ref.central_pos - 1
    r_ref_means, r_ref_sds, _, _ = ts.get_ref_from_seq(genome_seq, std_ref)
    # trim genome seq to match model-able positions
    genome_seq = genome_seq[std_ref.central_pos:-dnstrm_bases]
    seq_len = len(genome_seq)
    if genome_loc.Strand == '+':
        genome_loc = genome_loc._replace(
            Start=genome_loc.Start + std_ref.central_pos)
    else:
        genome_loc = genome_loc._replace(Start=genome_loc.Start + dnstrm_bases)

    # for short reads, just search the whole read with an appropriate bandwidth
    if (event_means.shape[0] < start_bandwidth + start_seq_window or
        seq_len < start_seq_window):
        seq_events = get_short_read_event_mapping(
            event_means, r_ref_means, r_ref_sds, skip_pen, stay_pen,
            z_shift, reg_id=reg_id, debug_fps=debug_fps)
        seq_segs = valid_cpts[seq_events]
        read_start_rel_to_raw = seq_segs[0]
        seq_segs = seq_segs - read_start_rel_to_raw

        return (seq_segs, r_ref_means, r_ref_sds, read_start_rel_to_raw,
                genome_seq, genome_loc)

    # identify the start and end of the read within the signal using a larger
    # bandwidth
    mapped_start, events_per_base = get_mapping_start(
        event_means, r_ref_means, r_ref_sds, skip_pen, stay_pen, z_shift,
        start_seq_window, start_bandwidth, norm_signal, valid_cpts,
        score_thresh, min_obs_per_base, reg_id=reg_id,
        debug_fps=debug_fps)

    # get number of events to clip and how far into the events the
    # discovered start is located
    half_bandwidth = bandwidth // 2
    if mapped_start < half_bandwidth:
        events_start_clip = 0
        mapped_start_offset = mapped_start
    else:
        events_start_clip = mapped_start - half_bandwidth
        mapped_start_offset = half_bandwidth

    # process long enough reads that start too far into read for normal
    # adaptive processing just as with short reads
    if (event_means.shape[0] - mapped_start_offset -
        events_start_clip < bandwidth):
        seq_events = get_short_read_event_mapping(
            event_means, r_ref_means, r_ref_sds, skip_pen, stay_pen,
            z_shift, reg_id=reg_id, debug_fps=debug_fps)
        seq_segs = valid_cpts[seq_events]
        read_start_rel_to_raw = seq_segs[0]
        seq_segs = seq_segs - read_start_rel_to_raw

        return (seq_segs, r_ref_means, r_ref_sds, read_start_rel_to_raw,
                genome_seq, genome_loc)

    # now find full sequence to events path using a smaller bandwidth
    event_means = event_means[events_start_clip:]
    valid_cpts = valid_cpts[events_start_clip:]
    (start_fwd_pass, start_fwd_pass_move,
     start_event_starts, start_z_scores) = get_masked_start_fwd_pass(
         event_means, r_ref_means, r_ref_sds,
         mapped_start_offset, skip_pen, stay_pen, z_shift, bandwidth,
         events_per_base, reg_id=reg_id, debug_fps=debug_fps)
    start_seq_len = start_event_starts.shape[0]
    fwd_pass = np.empty((seq_len+1, bandwidth), dtype=np.float64)
    fwd_pass[:start_seq_len+1] = start_fwd_pass
    fwd_pass_move = np.empty((seq_len+1, bandwidth), dtype=np.int64)
    fwd_pass_move[:start_seq_len+1] = start_fwd_pass_move
    band_event_starts = np.empty((seq_len,), dtype=np.int64)
    band_event_starts[:start_seq_len] = start_event_starts
    #fwd_pass[start_seq_len+1:,:] = np.NAN
    #fwd_pass_move[start_seq_len+1:,:] = np.NAN
    #band_event_starts[start_seq_len:] = np.NAN

    if _DEBUG_FULL or _DEBUG_MIDDLE:
        rest_z_scores = c_adaptive_banded_forward_pass(
            fwd_pass, fwd_pass_move, band_event_starts, event_means,
            r_ref_means, r_ref_sds, z_shift, skip_pen, stay_pen,
            start_seq_len, MASK_FILL_Z_SCORE, True)
        shifted_z_scores = np.empty((seq_len, bandwidth), dtype=np.float64)
        shifted_z_scores[:start_seq_len] = start_z_scores
        shifted_z_scores[start_seq_len:] = rest_z_scores
    else:
        c_adaptive_banded_forward_pass(
            fwd_pass, fwd_pass_move, band_event_starts, event_means,
            r_ref_means, r_ref_sds, z_shift, skip_pen, stay_pen,
            start_seq_len, MASK_FILL_Z_SCORE)

    top_max_pos = np.argmax(fwd_pass[-1,:])
    if _DEBUG_FULL:
        _write_full_debug(fwd_pass_move, band_event_starts, top_max_pos,
                         shifted_z_scores, debug_fps[2], debug_fps[3], reg_id,
                         fwd_pass[-1,top_max_pos])
    if _DEBUG_MIDDLE:
        _write_middle_debug(shifted_z_scores, fwd_pass, band_event_starts,
                           debug_fps[0], reg_id)
        _write_tb_debug(fwd_pass_move, band_event_starts, top_max_pos,
                       debug_fps[1], reg_id)

    read_tb = c_banded_traceback(
        fwd_pass_move, band_event_starts, top_max_pos, band_boundary_thresh)

    start_trim_i = 0
    while read_tb[start_trim_i] < 0:
        read_tb[start_trim_i] = 0
        start_trim_i += 1
    end_trim_i = 1
    events_len = event_means.shape[0]
    while read_tb[-end_trim_i] > events_len:
        read_tb[-end_trim_i] = events_len
        end_trim_i += 1

    seq_segs = valid_cpts[read_tb]
    read_start_rel_to_raw = seq_segs[0]
    seq_segs = seq_segs - read_start_rel_to_raw

    return (seq_segs, r_ref_means, r_ref_sds, read_start_rel_to_raw,
            genome_seq, genome_loc)


######################################
########## Re-squiggle Read ##########
######################################

def resquiggle_read(
        all_raw_signal, channel_info, genome_seq, genome_loc,
        align_info, std_ref, outlier_thresh, corr_grp,
        bio_samp_type, seg_params, sig_aln_params, obs_filter,
        fast5_fn=None, max_raw_cpts=MAX_RAW_CPTS,
        min_event_to_seq_ratio=MIN_EVENT_TO_SEQ_RATIO, skip_index=False,
        reg_id=None, debug_fps=None, const_scale=None):
    """
    Perform banded dynamic programming sequence to event alignment for this read

    :param all_raw_signal: Vector containing raw (DAC) current signal values
    :param channel_info: Channel info containing info for signal normalization
    :param fast5_fn: Relative path to filename for index creation
    :param genome_seq: Genomic sequence for this read
    :param genome_loc: Mapped genomic location named tuple for this read
    :param align_info: A alignInfo named tuple for this read
    :param std_ref: A TomboModel object
    :param outlier_thresh: Outlier threshold for raw signal normalization
    :param bc_grp: The basecalled read group to analyze
    :param corr_grp: The tombo corrected group to write results
    :param bio_samp_type: Biological sample type (either 'DNA' or 'RNA' or
        None to determine from read)
    :param seg_params: 3 segmenation parameters (mean_obs_per_event,
        running_stat_width and min_obs_per_base)
    :param sig_aln_params: Signal align parameters (match_evalue, skip_pen
        and bandwidth)
    :param obs_filter: Obervations per base filter to apply for filtered slot
        in FAST5
    """
    # flip raw signal for re-squiggling
    is_rna = bio_samp_type == 'RNA'
    if is_rna:
        all_raw_signal = all_raw_signal[::-1]

    if sig_aln_params is None:
        match_evalue, skip_pen, bandwidth, score_thresh = ALGN_PARAMS_TABLE[
            bio_samp_type]
    else:
        # unpack signal alignment parameters
        match_evalue, skip_pen, bandwidth, score_thresh = sig_aln_params
        bandwidth = int(bandwidth)
    z_shift, stay_pen = ts.get_dynamic_prog_params(match_evalue)

    if seg_params is None:
        (running_stat_width, min_obs_per_base,
         mean_obs_per_event) = SEG_PARAMS_TABLE[bio_samp_type]
    else:
        (running_stat_width, min_obs_per_base, mean_obs_per_event) = seg_params

    # compute number of events to find
    # ensure at least a minimal number of events per mapped sequence are found
    num_events = max(all_raw_signal.shape[0] // mean_obs_per_event,
                     int(len(genome_seq) * min_event_to_seq_ratio))
    # ensure that there isn't *far* too much signal for the mapped sequence
    # i.e. one adaptive bandwidth per base is too much to find a good mapping
    if num_events / bandwidth > len(genome_seq):
        raise NotImplementedError('Too much raw signal for mapped sequence')
    # normalize signal
    if const_scale is not None:
        norm_signal, scale_values = th.normalize_raw_signal(
            all_raw_signal, 0, all_raw_signal.shape[0],
            'median_const_scale', channel_info, outlier_thresh,
            const_scale=const_scale)
    else:
        norm_signal, scale_values = th.normalize_raw_signal(
            all_raw_signal, 0, all_raw_signal.shape[0],
            'median', channel_info, outlier_thresh)

    (segs, r_ref_means, r_ref_sds, read_start_rel_to_raw,
     genome_seq, genome_loc) = find_adaptive_base_assignment(
         norm_signal, running_stat_width, min_obs_per_base, num_events, std_ref,
         genome_seq, genome_loc, skip_pen, stay_pen, z_shift, bandwidth, is_rna,
         score_thresh, reg_id=reg_id, debug_fps=debug_fps)
    norm_signal = norm_signal[read_start_rel_to_raw:
                              read_start_rel_to_raw + segs[-1]]

    # identify all stretches of genomic deletions within del_fix_window
    # to be fixed.
    segs = get_model_fit_segs(
        segs, norm_signal, r_ref_means, r_ref_sds,
        min_obs_per_base, max_raw_cpts)

    if get_read_seg_score(
            norm_signal, segs, r_ref_means, r_ref_sds) > score_thresh:
        raise NotImplementedError('Poor raw to expected signal matching')
    if segs.shape[0] != len(genome_seq) + 1:
        raise ValueError('Aligned sequence does not match number ' +
                         'of segments produced')

    # Output for testing/visualization of re-squiggle
    if _DEBUG_PARAMS:
        _write_params_debug(
            norm_signal, segs, r_ref_means, r_ref_sds,
            running_stat_width, min_obs_per_base, mean_obs_per_event,
            match_evalue, skip_pen, bandwidth, fast5_fn)
    if _DEBUG_FIT:
        _write_fit_debug(
            norm_signal, segs, r_ref_means, r_ref_sds, genome_seq)

    return (genome_loc, read_start_rel_to_raw, segs, genome_seq,
            norm_signal, scale_values, corr_grp, align_info, is_rna)


#######################################
########## Genomic Alignment ##########
#######################################

def get_read_seq(fast5_data, bc_grp, bc_subgrp, bio_samp_type):
    """
    Extract the read sequence from the Fastq slot providing useful error messages
    """
    try:
        fastq_raw_value = fast5_data[
            '/Analyses/' + bc_grp + '/' + bc_subgrp + '/Fastq'].value
    except:
        raise NotImplementedError('Fastq slot not present in --basecall-group')

    # depending on how fastq data was stored it may already be encoded
    # as unicode, so this would fail.
    try:
        fastq_raw_value = fastq_raw_value.decode()
    except (TypeError, AttributeError):
        pass

    read_seq = fastq_raw_value.split('\n')[1]

    read_data = th.get_raw_read_slot(fast5_data)

    # looks like read_id attribute has been removed in some files and attribute
    # is not really necessary for tombo
    try:
        read_id = read_data.attrs['read_id']
    except:
        try:
            read_id = unicode(read_data.attrs['read_num'])
        except:
            read_id = unicode(np.random.randint(1000000000))

    try:
        if bio_samp_type is None:
            bio_samp_type = 'RNA' if th.is_read_rna(fast5_data) else 'DNA'
    except:
        raise NotImplementedError('Cannot determine whether read is DNA or RNA')
    if bio_samp_type == 'RNA':
        read_seq = th.rev_transcribe(read_seq)

    return read_seq, read_id, bio_samp_type

def map_read(fast5_data, bc_grp, bc_subgrp, corr_grp, aligner, bio_samp_type,
             map_thr_buf, genome_index):
    read_seq, read_id, bio_samp_type = get_read_seq(
        fast5_data, bc_grp, bc_subgrp, bio_samp_type)
    try:
        alignment = next(aligner.map(str(read_seq), buf=map_thr_buf))
    except StopIteration:
        raise NotImplementedError('Alignment not produced')

    chrm = alignment.ctg
    # subtract one to put into 0-based index
    ref_start = alignment.r_st
    ref_end = alignment.r_en
    strand = '+' if alignment.strand == 1 else '-'
    num_match = alignment.mlen
    num_ins, num_del, num_aligned = 0, 0 ,0
    for op_len, op in alignment.cigar:
        if op == 1: num_ins += op_len
        elif op in (2,3): num_del += op_len
        elif op in (0,7,8): num_aligned += op_len
        elif op == 6: pass
        else:
            # soft and hard clipping are not reported in the
            # mappy cigar
            raise NotImplementedError('Invalid cigar operation')
    if strand == '+':
        start_clipped_bases = alignment.q_st
        end_clipped_bases = len(read_seq) - alignment.q_en
    else:
        start_clipped_bases = len(read_seq) - alignment.q_en
        end_clipped_bases = alignment.q_st

    # current version of mappy has a bug in the sequence extraction for
    # long chromosomes, this line can be uncommented once the issue has been
    # resolved https://github.com/lh3/minimap2/issues/126
    #genome_seq = aligner.seq(chrm, ref_start, ref_end)
    genome_seq = genome_index.get_seq(chrm, ref_start, ref_end)
    if strand == '-':
        genome_seq = th.rev_comp(genome_seq)
    assert len(genome_seq) == ref_end - ref_start, (
        'Discordant mapped position and sequence')
    align_info = th.alignInfo(
        read_id, bc_subgrp, start_clipped_bases, end_clipped_bases,
        num_ins, num_del, num_match, num_aligned - num_match)
    genome_loc = th.genomeLoc(ref_start, strand, chrm)

    return genome_seq, genome_loc, align_info, bio_samp_type


########################################
########## Re-squiggle Worker ##########
########################################

def _resquiggle_worker(
        rsqgl_conns, std_ref, outlier_thresh, corr_grp, bio_samp_type,
        seg_params, sig_aln_params, obs_filter, skip_index, const_scale):
    debug_fps = None
    if _DEBUG_MIDDLE or _DEBUG_FULL:
        debug_fps = _open_debug_fps()

    while len(rsqgl_conns) > 0:
        # get next active connection or wait for one to be ready
        try:
            conn_num, rsqgl_conn = next(
                (conn_num, rsqgl_conn)
                for conn_num, rsqgl_conn in enumerate(rsqgl_conns)
                if rsqgl_conn.poll())
        except StopIteration:
            sleep(0.1)
            continue

        try:
            # get new mapping results
            map_info = rsqgl_conn.recv()
            # None signals the connected thread has finished the reads queue
            if map_info is None:
                del rsqgl_conns[conn_num]
                continue

            (all_raw_signal, channel_info, fast5_fn, genome_seq, genome_loc,
             align_info, bio_samp_type, reg_id) = map_info

            rsqgl_data = resquiggle_read(
                all_raw_signal, channel_info, genome_seq, genome_loc,
                align_info, std_ref, outlier_thresh, corr_grp, bio_samp_type,
                seg_params, sig_aln_params, obs_filter,
                fast5_fn=fast5_fn, skip_index=skip_index, reg_id=reg_id,
                debug_fps=debug_fps, const_scale=const_scale)
        except Exception as e:
            # uncomment to identify mysterious errors
            #raise
            rsqgl_conn.send([True, [unicode(e), fast5_fn]])
            continue

        rsqgl_conn.send([False, rsqgl_data])

    return

if _PROFILE_RSQGL:
    _resquiggle_wrapper = _resquiggle_worker
    def _resquiggle_worker(*args):
        import cProfile
        cProfile.runctx('_resquiggle_wrapper(*args)', globals(), locals(),
                        filename='resquiggle_main.prof')
        return
def _resquiggle_run_read(
        fast5_data, failed_reads_q, bc_subgrps, bc_grp, corr_grp, aligner,
        bio_samp_type, map_thr_buf, fast5_fn, num_processed, map_conn,
        outlier_thresh, compute_sd, obs_filter, index_q, genome_index):
    try:
        # extract channel and raw data for this read
        channel_info = th.get_channel_info(fast5_data)
        all_raw_signal = th.get_raw_read_slot(fast5_data)['Signal'].value
    except:
        failed_reads_q.put(
            ('Channel or raw signal information not found in FAST5 file',
             fast5_fn))

    for bc_subgrp in bc_subgrps:
        try:
            # TODO warn if reads appear to switch bio sample type
            genome_seq, genome_loc, align_info, bio_samp_type = map_read(
                fast5_data, bc_grp, bc_subgrp, corr_grp, aligner,
                bio_samp_type, map_thr_buf, genome_index)
            map_conn.send([
                all_raw_signal, channel_info, fast5_fn, genome_seq,
                genome_loc, align_info, bio_samp_type, num_processed])
            # wait until re-squiggle returns
            read_failed, rsqgl_data = map_conn.recv()
            if read_failed:
                failed_reads_q.put((
                    rsqgl_data[0], bc_subgrp + ':::' + rsqgl_data[1]))
                continue

            # unpack data needed to write new event data
            (genome_loc, read_start_rel_to_raw, segs, genome_seq,
             norm_signal, scale_values, corr_grp, align_info,
             is_rna) = rsqgl_data
            if not _DRY_RUN:
                # write re-squiggle event assignment to the read FAST5 file
                th.write_new_fast5_group(
                    fast5_data, genome_loc, read_start_rel_to_raw, segs,
                    genome_seq, norm_signal, scale_values, corr_grp,
                    align_info.Subgroup, 'median', outlier_thresh,
                    compute_sd, align_info=align_info, rna=is_rna)

            if index_q is not None:
                index_data = th.prep_index_data(
                    fast5_fn, genome_loc, read_start_rel_to_raw,
                    segs, corr_grp, align_info.Subgroup, is_rna, obs_filter)

                index_q.put(index_data)
                if index_data[1][6]:
                    failed_reads_q.put((
                        'Read filtered by observation per base ' +
                        'thresholds (revert with `tombo clear_filters`)',
                        bc_subgrp + ':::' + fast5_fn))
        except Exception as e:
            # uncomment to identify mysterious errors
            #raise
            try:
                th.write_error_status(
                    fast5_fn, corr_grp, bc_subgrp, unicode(e))
            except:
                pass
            failed_reads_q.put((
                unicode(e), bc_subgrp + ':::' + fast5_fn))

    return

def _resquiggle_mappy_thread_worker(
        fast5_q, progress_q, failed_reads_q, index_q, bc_grp, bc_subgrps,
        corr_grp, aligner, outlier_thresh, compute_sd, sig_aln_params,
        obs_filter, const_scale, bio_samp_type, seg_params,
        overwrite, map_conn, genome_fn):
    # get mappy aligner thread buffer
    map_thr_buf = mappy.ThreadBuffer()
    genome_index = th.Fasta(genome_fn)

    num_processed = 0
    while True:
        try:
            fast5_fn = fast5_q.get(block=False)
        except queue.Empty:
            # signal that all reads have been processed to child process
            map_conn.send(None)
            # update with all reads processed from this thread
            progress_q.put(num_processed % PROC_UPDATE_INTERVAL)
            break

        num_processed += 1
        if num_processed % PROC_UPDATE_INTERVAL == 0:
            progress_q.put(PROC_UPDATE_INTERVAL)

        if _DRY_RUN:
            prep_result = h5py.File(fast5_fn, 'r')
        else:
            # prep the fast5 file for writing
            prep_result = th.prep_fast5(
                fast5_fn, corr_grp, overwrite, True, bc_grp, return_fp=True)
        if isinstance(prep_result, h5py.File):
            fast5_data = prep_result
        else:
            failed_reads_q.put(prep_result)
            continue

        try:
            _resquiggle_run_read(
                fast5_data, failed_reads_q, bc_subgrps, bc_grp, corr_grp,
                aligner, bio_samp_type, map_thr_buf, fast5_fn,
                num_processed, map_conn, outlier_thresh, compute_sd,
                obs_filter, index_q, genome_index)
        finally:
            try:
                fast5_data.close()
            except:
                failed_reads_q.put(('Error closing fast5 file', fast5_fn))

    return


###########################################
########## Re-squiggle All Reads ##########
###########################################

def load_minimap_index(genome_fn, mm_index):
    if mm_index:
        aligner = mappy.Aligner(str(mm_index), preset=str('map-ont'))
    else:
        aligner = mappy.Aligner(str(genome_fn), preset=str('map-ont'))

    return aligner

def resquiggle_all_reads(
        fast5_fns, aligner, bc_grp, bc_subgrps, corr_grp, std_ref,
        bio_samp_type, outlier_thresh, overwrite, num_ps, threads_per_proc,
        compute_sd, skip_index, sig_aln_params, obs_filter, const_scale,
        seg_params, genome_fn):
    """
    Perform genomic alignment and re-squiggle algorithm
    """
    fast5_q = queue.Queue()
    failed_reads_q = queue.Queue()
    index_q = queue.Queue() if not skip_index else None
    progress_q = queue.Queue()
    for fast5_fn in fast5_fns:
        fast5_q.put(fast5_fn)

    # open all multiprocessing pipes and queues before threading
    # as opening threads before all process are open seems to cause
    # a deadlock when some processes are started.
    # starting all multiprocess objects seems to fix this.
    map_conns = []
    for _ in range(num_ps):
        proc_rsqgl_conns = []
        for _ in range(threads_per_proc):
            # open mp pipe to communicate with re-squiggle process
            map_conn, rsqgl_conn = Pipe()
            map_conns.append(map_conn)
            proc_rsqgl_conns.append(rsqgl_conn)
        # open re-squiggle process to void intensive processing hitting the GIL
        rsqgl_args = (
            proc_rsqgl_conns, std_ref, outlier_thresh, corr_grp, bio_samp_type,
            seg_params, sig_aln_params, obs_filter, index_q is None, const_scale)
        rsqgl_process = Process(target=_resquiggle_worker, args=rsqgl_args,
                                daemon=True)
        rsqgl_process.start()

    # now open mapping thread for each map connection created above
    resquiggle_ts = []
    for map_conn in map_conns:
        map_args = (fast5_q, progress_q, failed_reads_q, index_q, bc_grp,
                    bc_subgrps, corr_grp, aligner, outlier_thresh, compute_sd,
                    sig_aln_params, obs_filter, const_scale, bio_samp_type,
                    seg_params, overwrite, map_conn, genome_fn)
        t = threading.Thread(target=_resquiggle_mappy_thread_worker,
                             args=map_args)
        t.start()
        resquiggle_ts.append(t)

    if VERBOSE: sys.stderr.write(
            'Correcting ' + unicode(len(fast5_fns)) + ' files with ' +
            unicode(len(bc_subgrps)) + ' subgroup(s)/read(s) ' +
            'each (Will print a dot for each ' + unicode(PROGRESS_INTERVAL) +
            ' reads completed).\n')
    tot_num_rec_proc = 0
    failed_reads = defaultdict(list)
    all_index_data = []
    # note that this thread is counted so there will be 1 thread left when
    # all children are done
    while threading.active_count() > 1:
        try:
            errorType, fn = failed_reads_q.get(block=False)
            failed_reads[errorType].append(fn)
        except queue.Empty:
            try:
                num_rec_proc = progress_q.get(block=False)
                num_int_proc = (
                    ((tot_num_rec_proc + num_rec_proc) // PROGRESS_INTERVAL) -
                    (tot_num_rec_proc // PROGRESS_INTERVAL))
                if num_int_proc > 0:
                    sys.stderr.write('.' * num_int_proc)
                    sys.stderr.flush()
                tot_num_rec_proc += num_rec_proc
            except queue.Empty:
                if index_q is not None:
                    try:
                        r_index_data = index_q.get(block=False)
                        all_index_data.append(r_index_data)
                    except queue.Empty:
                        sleep(1)
                        continue

    # empty any entries left in queue after processes have finished
    while not failed_reads_q.empty():
        errorType, fn = failed_reads_q.get(block=False)
        failed_reads[errorType].append(fn)
    if index_q is not None:
        while not index_q.empty():
            r_index_data = index_q.get(block=False)
            all_index_data.append(r_index_data)

    # join all threads back with main thread
    for t in resquiggle_ts:
        t.join()

    # print newline after read progress dots
    if VERBOSE: sys.stderr.write('\n')

    return dict(failed_reads), all_index_data

def parse_files(args):
    if VERBOSE: sys.stderr.write('Getting file list.\n')
    try:
        if not os.path.isdir(args.fast5_basedir):
            th._error_message_and_exit(
                'Provided [fast5-basedir] is not a directory.')
        fast5_basedir = (
            args.fast5_basedir if args.fast5_basedir.endswith('/') else
            args.fast5_basedir + '/')
        if args.skip_index:
            index_fn = None
        else:
            index_fn = th.get_index_fn(fast5_basedir, args.corrected_group)
            if os.path.exists(index_fn): os.remove(index_fn)

        files, lock_fns = th.get_files_list_and_lock_dirs(
            fast5_basedir, args.ignore_read_locks)
    except OSError:
        th._error_message_and_exit(
            'Reads base directory, a sub-directory ' +
            'or an old (hidden) index file does not appear to be ' +
            'accessible. Check directory permissions.')
    if len(files) < 1:
        th.clear_tombo_locks(lock_fns)
        th._error_message_and_exit(
            'No files identified in the specified ' +
            'directory or within immediate subdirectories.')

    if not th.reads_contain_basecalls(
            files, args.basecall_group, num_reads=1000):
        th.clear_tombo_locks(lock_fns)
        th._error_message_and_exit(
            'Reads do not to contain basecalls. Check --basecall-group option ' +
            'if basecalls are stored in non-standard location or use ' +
            '`tombo annotate_raw_with_fastqs` to add basecalls from FASTQ ' +
            'files to raw FAST5 files.')

    return files, fast5_basedir, index_fn, lock_fns


###################################
########## Main Function ##########
###################################

def resquiggle_main(args):
    """
    Main method for resquiggle
    """
    global VERBOSE
    VERBOSE = not args.quiet
    th.VERBOSE = VERBOSE
    ts.VERBOSE = VERBOSE

    if args.basecall_group == args.corrected_group:
        th._error_message_and_exit(
            '--basecall-group and --corrected-group must ' +
            'be different.')

    # check simple arguments for validity first
    outlier_thresh = args.outlier_threshold if (
        args.outlier_threshold > 0) else None
    obs_filter = th.parse_obs_filter(args.obs_per_base_filter) \
                 if 'obs_per_base_filter' in args else None

    if VERBOSE: sys.stderr.write('Loading minimap2 reference.\n')
    # to be enabled when mappy genome sequence extraction bug is fixed
    #aligner = mappy.Aligner(str(args.reference_fasta), preset=str('map-ont'))
    aligner = load_minimap_index(args.genome_fasta, args.minimap2_index)
    if not aligner:
        th._error_message_and_exit(
            'Failed to load reference genome FASTA for mapping.')

    files, fast5_basedir, index_fn, lock_fns = parse_files(args)

    try:
        tb_model_fn = args.tombo_model_filename
        bio_samp_type = args.bio_sample_type
        if tb_model_fn is None:
            tb_model_fn, bio_samp_type = ts.get_default_standard_ref_from_files(
                files, bio_samp_type)
        if not os.path.exists(tb_model_fn):
            th._error_message_and_exit('Invalid tombo model file provided.')
        # parse tombo model
        std_ref = ts.TomboModel(tb_model_fn)

        const_scale = None
        if args.fixed_scale is not None:
            const_scale = args.fixed_scale
        elif not args.fit_scale_per_read:
            const_scale = th.estimate_global_scale(files)

        failed_reads, all_index_data = resquiggle_all_reads(
            files, aligner, args.basecall_group, args.basecall_subgroups,
            args.corrected_group, std_ref, bio_samp_type, outlier_thresh,
            args.overwrite, args.processes, args.threads_per_process,
            args.include_event_stdev, args.skip_index,
            args.signal_align_parameters, obs_filter, const_scale,
            args.segmentation_parameters, args.genome_fasta)
    finally:
        th.clear_tombo_locks(lock_fns)

    if not args.skip_index:
        th.write_index_file(all_index_data, index_fn, fast5_basedir)
    fail_summary = [(err, len(fns)) for err, fns in failed_reads.items()]
    if len(fail_summary) > 0:
        total_num_failed = sum(map(itemgetter(1), fail_summary))
        sys.stderr.write('Failed reads summary (' + unicode(total_num_failed) +
                         ' total failed):\n' + '\n'.join(
                             "\t" + err + " :\t" + unicode(n_fns)
                             for err, n_fns in sorted(fail_summary)) + '\n')
    else:
        sys.stderr.write('All reads successfully re-squiggled!\n')
    if args.failed_reads_filename is not None:
        with io.open(args.failed_reads_filename, 'wt') as fp:
            fp.write('\n'.join((
                err + '\t' + ', '.join(fns)
                for err, fns in failed_reads.items())) + '\n')

    return

def _args_and_main():
    import _option_parsers
    resquiggle_main(
        _option_parsers.get_resquiggle_parser().parse_args())
    return

if __name__ == '__main__':
    _args_and_main()



from __future__ import unicode_literals

###############################
##### Model Name Defaults #####
###############################

# default model names
STANDARD_MODELS = {
    'DNA':'tombo.DNA.model',
    'RNA':'tombo.RNA.200mV.model',
}
ALTERNATE_MODELS = {
    'DNA_5mC':'tombo.DNA.5mC.model',
    'DNA_6mA':'tombo.DNA.6mA.model',
}


################################
##### Re-squiggle Defaults #####
################################

# table containing default segmentation parameters for different sample types
#   1) running neighboring window width for segmentation scoring
#   2) minimum observations per genomic base
#   3) mean number of observations per event during segmentation
SEG_PARAMS_TABLE = {
    'RNA':(8, 4, 10),
    'DNA':(5, 3, 5),
}

# table containing default signal to sequence assignment parameters
# for different sample types
#   1) expected value for matching event to sequence
#   2) penalty for skipped sequence position
#   3) adaptive bandwidth
#   4) signal segmentation mean half-normal score threshold
ALGN_PARAMS_TABLE = {
    'RNA':(4, 10, 1400, 2.0),
    'DNA':(4.2, 4.2, 1200, 1.75),
}

# factor of extra raw signal above minimum to add around skipped bases for
# raw signal segment detection
EXTRA_SIG_FACTOR = 1.1

MASK_FILL_Z_SCORE = -10
MASK_BASES = 50

START_BANDWIDTH = 5000
START_SEQ_WINDOW = 500
BAND_BOUNDARY_THRESH = 5

DEL_FIX_WINDOW = 2
MAX_DEL_FIX_WINDOW = 8
MAX_RAW_CPTS = 200
MIN_EVENT_TO_SEQ_RATIO = 1.1


############################
##### Testing Defaults #####
############################

LLR_THRESH = 0.0
SAMP_COMP_THRESH = 0.1
DE_NOVO_THRESH = 0.5


#####################################
##### Model Estimation Defaults #####
#####################################

ALT_EST_BATCH = 1000
MAX_KMER_OBS = 10000
MIN_KMER_OBS_TO_EST = 50
KERNEL_DENSITY_RANGE = (-5,5)


##########################
##### Misc. Defaults #####
##########################

SMALLEST_PVAL = 1e-50

# got quantiles from analysis of stability after shift-only normalization
ROBUST_QUANTS = (46.5, 53.5)

# minimum standard deviation for a genomic position when estimating spread
# from a sample
MIN_POSITION_SD = 0.01

# number of points at which to estimate the k-mer siganl densities
NUM_DENS_POINTS = 500

# number of reads to estimate global scale parameter
NUM_READS_FOR_SCALE = 1000

# number of points to plot in the ROC curve plotting command
ROC_PLOT_POINTS = 1000



from __future__ import division, unicode_literals, absolute_import

from builtins import int, range, dict, map, zip

import io
import sys

import numpy as np

from collections import defaultdict

if sys.version_info[0] > 2:
    unicode = str

# import tombo functions
from . import tombo_stats as ts
from . import tombo_helper as th

from ._default_parameters import SMALLEST_PVAL

VERBOSE = False

WIG_HEADER='track type=wiggle_0 name="{0}_{1}_{2}{3}" ' + \
    'description="{0} {1} {2}{4}"\n'
GROUP1_NAME='sample'
GROUP2_NAME='control'


########################
###### WIG Output ######
########################

def _write_wiggle(wig_base, group_text, data_values, type_name,
                 filter_zeros=False):
    group_w_dot = '' if group_text == '' else '.' + group_text
    group_w_us = '' if group_text == '' else '_' + group_text
    group_w_space = '' if group_text == '' else ' ' + group_text
    plus_wig_fp = io.open(
        wig_base + '.' + type_name + group_w_dot + '.plus.wig', 'wt')
    minus_wig_fp = io.open(
        wig_base + '.' + type_name + group_w_dot + '.minus.wig', 'wt')
    plus_wig_fp.write(WIG_HEADER.format(
        wig_base, type_name, 'fwd_strand', group_w_us, group_w_space))
    minus_wig_fp.write(WIG_HEADER.format(
        wig_base, type_name, 'rev_strand', group_w_us, group_w_space))
    for (chrm, strand), cs_values in data_values.items():
        wig_fp = plus_wig_fp if strand == '+' else minus_wig_fp
        wig_fp.write("variableStep chrom={} span=1\n".format(chrm))
        wig_fp.write('\n'.join([
            unicode(int(pos) + 1) + " " + unicode(round(val, 4))
            for pos, val in enumerate(cs_values)
            if not (np.isnan(val) or (
                    filter_zeros and np.equal(val, 0.0)))]) + '\n')

    plus_wig_fp.close()
    minus_wig_fp.close()

    return

def write_frac_wigs(all_stats, wig_base, do_frac, do_damp):
    if VERBOSE: sys.stderr.write('Parsing statistics.\n')
    if do_frac:
        all_frac = {}
    if do_damp:
        all_damp_frac = {}

    curr_chrm, curr_strand, curr_poss, curr_fracs, curr_damp_fracs = (
        None, None, [], [], [])
    for chrm, strand, pos, frac, damp_frac in all_stats.iter_fracs():
        if chrm != curr_chrm or strand != curr_strand:
            if len(curr_poss) > 0:
                cs_max_pos = max(curr_poss)
                # store current data
                if do_frac:
                    cs_fracs = np.empty(cs_max_pos + 1)
                    cs_fracs[:] = np.nan
                    np.put(cs_fracs, curr_poss, curr_fracs)
                    all_frac[(curr_chrm, curr_strand)] = cs_fracs
                if do_damp:
                    cs_damps = np.empty(cs_max_pos + 1)
                    cs_damps[:] = np.nan
                    np.put(cs_damps, curr_poss, curr_damp_fracs)
                    all_damp_frac[(curr_chrm, curr_strand)] = cs_damps

            # set new chrm and strand and empty lists
            curr_chrm, curr_strand = chrm, strand
            curr_poss, curr_fracs, curr_damp_frac = [], [], []

        # store position statistics
        curr_poss.append(pos)
        if do_frac:
            curr_fracs.append(1 - frac)
        if do_damp:
            curr_damp_fracs.append(1 - damp_frac)

    # tabulate and store last chrm and strand
    if len(curr_poss) > 0:
        cs_max_pos = max(curr_poss)
        # store current data
        if do_frac:
            cs_fracs = np.empty(cs_max_pos + 1)
            cs_fracs[:] = np.nan
            np.put(cs_fracs, curr_poss, curr_fracs)
            all_frac[(curr_chrm, curr_strand)] = cs_fracs
        if do_damp:
            cs_damps = np.empty(cs_max_pos + 1)
            cs_damps[:] = np.nan
            np.put(cs_damps, curr_poss, curr_damp_fracs)
            all_damp_frac[(curr_chrm, curr_strand)] = cs_damps

    if VERBOSE: sys.stderr.write('Writing fraction wigs.\n')
    if do_frac:
        _write_wiggle(wig_base, '', all_frac, 'fraction_modified_reads')
    if do_damp:
        _write_wiggle(wig_base, '', all_damp_frac,
                      'dampened_fraction_modified_reads')

    return

def write_length_wig(
        raw_read_coverage, chrm_sizes, wig_base, group_name):
    if VERBOSE: sys.stderr.write('Parsing events lengths.\n')
    base_lens = th.get_all_mean_lengths(raw_read_coverage, chrm_sizes)

    if VERBOSE: sys.stderr.write('Writing length wig.\n')
    _write_wiggle(wig_base, group_name, base_lens, 'dwell')

    return

def write_signal_sd_wig(
        raw_read_coverage, chrm_sizes, wig_base, group_name):
    if VERBOSE: sys.stderr.write('Parsing signal SDs.\n')
    base_sds = th.get_all_mean_stdev(raw_read_coverage, chrm_sizes)

    if VERBOSE: sys.stderr.write('Writing signal SD wig.\n')
    _write_wiggle(wig_base, group_name, base_sds, 'signalSd')

    return

def write_signal_and_diff_wigs(
        raw_read_coverage1, raw_read_coverage2, chrm_sizes,
        wig_base, group1_name, write_sig, write_diff):
    if VERBOSE: sys.stderr.write('Parsing mean base signals.\n')
    base_means1 = th.get_all_mean_levels(raw_read_coverage1, chrm_sizes)
    if raw_read_coverage2 is not None:
        base_means2 = th.get_all_mean_levels(raw_read_coverage2, chrm_sizes)

        if write_diff:
            if VERBOSE: sys.stderr.write(
                    'Calculating signal differences.\n')
            sig_diffs = {}
            for chrm, strand in [(c, s) for c in chrm_sizes
                                 for s in ('+', '-')]:
                # calculate difference and set no coverage
                # (nan) values to zero
                sig_diffs[(chrm, strand)] = (base_means1[(chrm, strand)] -
                                             base_means2[(chrm, strand)])
            if VERBOSE: sys.stderr.write('Writing differnce wig.\n')
            _write_wiggle(wig_base, '', sig_diffs, 'difference')
        if write_sig:
            if VERBOSE: sys.stderr.write('Writing signal wigs.\n')
            _write_wiggle(wig_base, GROUP2_NAME, base_means2, 'signal')

    if write_sig:
        _write_wiggle(wig_base, group1_name, base_means1, 'signal')

    return

def write_cov_wig(raw_read_coverage, wig_base, group_text):
    read_coverage = th.get_coverage(raw_read_coverage)

    if VERBOSE: sys.stderr.write('Writing coverage wig.\n')
    _write_wiggle(wig_base, group_text, read_coverage, 'coverage', True)

    return

def write_all_wiggles(
        f5_dirs1, f5_dirs2, corr_grp, bc_subgrps,
        stats_fn, wig_base, wig_types, cov_damp_counts):
    if f5_dirs1 is not None:
        raw_read_coverage1 = th.parse_fast5s(
            f5_dirs1, corr_grp, bc_subgrps)
        if len(raw_read_coverage1) == 0:
            th._error_message_and_exit(
                'No reads present in --fast5-basedirs.')

    group1_name = '' if f5_dirs2 is None else GROUP1_NAME
    if f5_dirs2 is not None:
        raw_read_coverage2 = th.parse_fast5s(
            f5_dirs2, corr_grp, bc_subgrps)
        chrm_sizes = th.get_chrm_sizes(
            raw_read_coverage1, raw_read_coverage2)

        if VERBOSE: sys.stderr.write('Writing wiggles.\n')
        if 'coverage' in wig_types:
            write_cov_wig(raw_read_coverage2, wig_base, GROUP2_NAME)
        if 'signal_sd' in wig_types:
            write_signal_sd_wig(
                raw_read_coverage2, chrm_sizes, wig_base, GROUP2_NAME)
        if 'dwell' in wig_types:
            write_length_wig(raw_read_coverage2, chrm_sizes,
                             wig_base, GROUP2_NAME)

        # need to do signal and difference call once either with or
        # w/o second set of files (unlike coverage, sds and length
        if 'signal' in wig_types or 'difference' in wig_types:
            write_signal_and_diff_wigs(
                raw_read_coverage1, raw_read_coverage2, chrm_sizes,
                wig_base, group1_name, 'signal' in wig_types,
                'difference' in wig_types)
    elif f5_dirs1 is not None:
        chrm_sizes = th.get_chrm_sizes(raw_read_coverage1)
        if VERBOSE: sys.stderr.write('Writing wiggles.\n')
        if 'signal' in wig_types:
            write_signal_and_diff_wigs(
                raw_read_coverage1, None, chrm_sizes, wig_base,
                group1_name, 'signal' in wig_types, False)

    if 'coverage' in wig_types:
        write_cov_wig(raw_read_coverage1, wig_base, group1_name)
    if 'signal_sd' in wig_types:
        write_signal_sd_wig(
            raw_read_coverage1, chrm_sizes, wig_base, group1_name)
    if 'dwell' in wig_types:
        write_length_wig(raw_read_coverage1, chrm_sizes, wig_base, group1_name)
    if any(wig_type in wig_types for wig_type in (
            'fraction', 'dampened_fraction')):
        if VERBOSE: sys.stderr.write('Loading statistics from file.\n')
        all_stats = ts.TomboStats(stats_fn)
        if 'dampened_fraction' in wig_types:
            all_stats.calc_damp_fraction(cov_damp_counts)
        all_stats.order_by_pos()
        write_frac_wigs(all_stats, wig_base, 'fraction' in wig_types,
                        'dampened_fraction' in wig_types)

    return


##########################
###### FASTA Output ######
##########################

def write_most_signif(
        f5_dirs, fasta_fn, num_regions, corr_grp, bc_subgrps, seqs_fn,
        num_bases, stats_fn, cov_damp_counts):
    if VERBOSE: sys.stderr.write('Loading statistics from file.\n')
    plot_intervals = ts.TomboStats(stats_fn).get_most_signif_regions(
        num_bases, num_regions, cov_damp_counts=cov_damp_counts)

    # get each regions sequence either from reads or fasta index
    if fasta_fn is None:
        raw_read_coverage = th.parse_fast5s(f5_dirs, corr_grp, bc_subgrps)
        all_reg_data = th.get_region_sequences(
            plot_intervals, raw_read_coverage)
    else:
        genome_index = th.Fasta(fasta_fn)
        all_reg_data = [
            int_i._replace(
                seq=genome_index.get_seq(int_i.chrm, int_i.start, int_i.end))
            for int_i in plot_intervals
            if int_i.chrm in genome_index.index.index]

    if VERBOSE: sys.stderr.write('Outputting region seqeuences.\n')
    with io.open(seqs_fn, 'wt') as seqs_fp:
        for int_i in all_reg_data:
            reg_seq = int_i.seq
            if int_i.strand == '-':
                reg_seq = th.rev_comp(reg_seq)
            seqs_fp.write('>{0}:{1:d}:{2} {3}\n{4}\n'.format(
                int_i.chrm, int(int_i.start + (num_bases // 2)),
                int_i.strand, int_i.reg_text, ''.join(reg_seq)))

    return


############################
###### Main functions ######
############################

def wiggle_main(args):
    global VERBOSE
    VERBOSE = not args.quiet
    th.VERBOSE = VERBOSE
    ts.VERBOSE = VERBOSE

    if (any(data_type in args.wiggle_types
            for data_type in ['signal', 'difference', 'coverage',
                              'signal_sd', 'dwell']) and
        args.fast5_basedirs is None):
        th._error_message_and_exit(
            'Must provide a fast5 basedir to output signal, difference, ' +
            'coverage, signal_sd and/or length wiggle files.')
    if (any(wig_type in args.wiggle_types for wig_type in (
            'fraction', 'dampened_fraction')) and
        args.statistics_filename is None):
        th._error_message_and_exit(
            'Must provide a statistics filename to output ' +
            'fraction wiggle files.')
    if ('difference' in args.wiggle_types and
        args.control_fast5_basedirs is None):
        th._error_message_and_exit(
            'Must provide two sets of FAST5s ' + \
            'to output difference wiggle files.')
    if (args.control_fast5_basedirs is not None and
        args.fast5_basedirs is None):
        th._error_message_and_exit(
            'Cannot provide a control FAST5 set of directories ' +
            'without a sample set of FAST5 directories.')
    if (args.coverage_dampen_counts is None and
        'dampened_fraction' in args.wiggle_types):
        th._error_message_and_exit(
            'Cannot compute dampened fractions without ' +
            '--coverage-dampened-counts values.')

    write_all_wiggles(
        args.fast5_basedirs, args.control_fast5_basedirs, args.corrected_group,
        args.basecall_subgroups, args.statistics_filename, args.wiggle_basename,
        args.wiggle_types, args.coverage_dampen_counts)

    return

def write_signif_diff_main(args):
    global VERBOSE
    VERBOSE = not args.quiet
    th.VERBOSE = VERBOSE
    ts.VERBOSE = VERBOSE

    if args.fast5_basedirs is None and args.genome_fasta is None:
        th._error_message_and_exit(
            'Must provide either FAST5 directory(ies) or a fasta file.')

    write_most_signif(
        args.fast5_basedirs, args.genome_fasta, args.num_regions,
        args.corrected_group, args.basecall_subgroups, args.sequences_filename,
        args.num_bases, args.statistics_filename, args.coverage_dampen_counts)

    return


if __name__ == '__main__':
    raise NotImplementedError(
        'This is a module. See commands with `tombo -h`')



import numpy as np
cimport numpy as np

DTYPE = np.float64
ctypedef np.float64_t DTYPE_t

DTYPE_INT = np.int64
ctypedef np.int64_t DTYPE_INT_t

from libc.math cimport log

cdef extern from "math.h":
    double sqrt(double m)

def c_mean_std(np.ndarray[DTYPE_t] values):
    """
    More efficient method to get both mean and standard deviation
    """
    cdef DTYPE_t v_mean, v_var
    cdef DTYPE_INT_t idx
    cdef DTYPE_INT_t v_len = values.shape[0]
    v_mean = 0
    for idx in range(v_len):
        v_mean += values[idx]
    v_mean /= v_len
    v_var = 0
    for idx in range(v_len):
        v_var += (values[idx] - v_mean)**2
    return v_mean, sqrt(v_var / v_len)

def c_new_mean_stds(np.ndarray[DTYPE_t] norm_signal not None,
                    np.ndarray[DTYPE_INT_t] new_segs not None):
    cdef DTYPE_INT_t n_segs = new_segs.shape[0] - 1
    cdef np.ndarray[DTYPE_t] means_arr = np.empty(n_segs, dtype=DTYPE)
    cdef np.ndarray[DTYPE_t] stds_arr = np.empty(n_segs, dtype=DTYPE)
    cdef DTYPE_t curr_sum, curr_var, seg_mean
    cdef DTYPE_INT_t idx, seg_idx, seg_len
    for idx in range(n_segs):
        seg_len = new_segs[idx + 1] - new_segs[idx]
        curr_sum = 0
        for seg_idx in range(new_segs[idx], new_segs[idx + 1]):
            curr_sum += norm_signal[seg_idx]
        seg_mean = curr_sum / seg_len
        means_arr[idx] = seg_mean
        curr_var = 0
        for seg_idx in range(new_segs[idx], new_segs[idx + 1]):
            curr_var += (norm_signal[seg_idx] - seg_mean)**2
        stds_arr[idx] = sqrt(curr_var / seg_len)
    return means_arr, stds_arr

def c_new_means(np.ndarray[DTYPE_t] norm_signal not None,
                np.ndarray[DTYPE_INT_t] new_segs not None):
    cdef DTYPE_INT_t n_segs = new_segs.shape[0] - 1
    cdef np.ndarray[DTYPE_t] means_arr = np.empty(n_segs, dtype=DTYPE)
    cdef DTYPE_t curr_sum
    cdef DTYPE_INT_t idx, seg_idx
    for idx in range(n_segs):
        curr_sum = 0
        for seg_idx in range(new_segs[idx], new_segs[idx + 1]):
            curr_sum += norm_signal[seg_idx]
        means_arr[idx] = curr_sum / (new_segs[idx + 1] - new_segs[idx])
    return means_arr

def c_apply_outlier_thresh(
        np.ndarray[DTYPE_t] raw_signal, DTYPE_t lower_lim, DTYPE_t upper_lim):
    cdef DTYPE_INT_t raw_size = raw_signal.shape[0]
    cdef np.ndarray[DTYPE_t] clipped_signal = np.empty(raw_size, dtype=DTYPE)
    cdef DTYPE_INT_t pos
    cdef DTYPE_t pos_sig
    for pos in range(raw_size):
        pos_sig = raw_signal[pos]
        if pos_sig > upper_lim:
            clipped_signal[pos] = upper_lim
        elif pos_sig < lower_lim:
            clipped_signal[pos] = lower_lim
        else:
            clipped_signal[pos] = pos_sig
    return clipped_signal

def c_valid_cpts_w_cap(
        np.ndarray[DTYPE_t] raw_signal, DTYPE_INT_t min_base_obs,
        DTYPE_INT_t running_stat_width, DTYPE_INT_t num_cpts):
    cdef np.ndarray[DTYPE_t] raw_cumsum = np.cumsum(
        np.concatenate([[0.0], raw_signal]))
    # get difference between all neighboring running_stat_width regions
    cdef np.ndarray[DTYPE_INT_t] candidate_poss = np.argsort(np.abs(
        (2 * raw_cumsum[running_stat_width:-running_stat_width]) -
        raw_cumsum[:-2*running_stat_width] -
        raw_cumsum[2*running_stat_width:])).astype(DTYPE_INT)[::-1]

    cdef np.ndarray[DTYPE_INT_t] cpts = np.empty(num_cpts, dtype=DTYPE_INT)
    cpts[0] = candidate_poss[0] + running_stat_width
    blacklist_pos = set(range(
        candidate_poss[0] - min_base_obs + 1, candidate_poss[0] + min_base_obs))
    cdef DTYPE_INT_t cand_pos
    cdef DTYPE_INT_t num_cands = candidate_poss.shape[0] - (
        2 * running_stat_width)
    cdef DTYPE_INT_t cand_idx = 1
    cdef DTYPE_INT_t added_cpts = 1
    while added_cpts < num_cpts:
        cand_pos = candidate_poss[cand_idx]
        if cand_pos not in blacklist_pos:
            cpts[added_cpts] = cand_pos + running_stat_width
            added_cpts += 1
            blacklist_pos.update(range(
                cand_pos - min_base_obs + 1, cand_pos + min_base_obs))
        cand_idx += 1
        if cand_idx >= num_cands:
            raise NotImplementedError, 'Fewer changepoints found than requested'

    return cpts

def c_valid_cpts(np.ndarray[DTYPE_t] raw_signal, DTYPE_INT_t min_base_obs,
                 DTYPE_INT_t running_stat_width):
    cdef np.ndarray[DTYPE_t] raw_cumsum = np.cumsum(
        np.concatenate([[0.0], raw_signal]))
    # get difference between all neighboring running_stat_width regions
    cdef np.ndarray[DTYPE_INT_t] candidate_poss = np.argsort(np.abs(
        (2 * raw_cumsum[running_stat_width:-running_stat_width]) -
        raw_cumsum[:-2*running_stat_width] -
        raw_cumsum[2*running_stat_width:])).astype(DTYPE_INT)[::-1]

    cpts = [candidate_poss[0]]
    blacklist_pos = set()
    cdef DTYPE_INT_t pos
    for pos in candidate_poss[1:]:
        if pos not in blacklist_pos:
            cpts.append(pos)
            blacklist_pos.update(range(
                pos-min_base_obs+1, pos+min_base_obs+1))

    return np.array(cpts) + running_stat_width

def c_valid_cpts_w_cap_t_test(
        np.ndarray[DTYPE_t] raw_signal, DTYPE_INT_t min_base_obs,
        DTYPE_INT_t running_stat_width, DTYPE_INT_t num_cpts):
    cdef DTYPE_INT_t pos, idx
    cdef DTYPE_t pos_diff, m1, m2, var1, var2
    cdef DTYPE_INT_t num_cands = raw_signal.shape[0] - (running_stat_width * 2)
    # note these will not actually be t-scores, but will be a monotonic transform
    # so the rank order will be the same
    cdef np.ndarray[DTYPE_t] t_scores = np.empty(num_cands, dtype=DTYPE)
    for pos in range(num_cands):
        # compute means
        m1 = 0
        for idx in range(running_stat_width):
            m1 += raw_signal[pos + idx]
        m1 /= running_stat_width
        m2 = 0
        for idx in range(running_stat_width):
            m2 += raw_signal[pos + running_stat_width + idx]
        m2 /= running_stat_width

        # compute sum of variances
        var1 = 0
        for idx in range(running_stat_width):
            pos_diff = raw_signal[pos + idx] - m1
            var1 += pos_diff * pos_diff
        var2 = 0
        for idx in range(running_stat_width):
            pos_diff = raw_signal[pos + running_stat_width + idx] - m2
            var2 += pos_diff * pos_diff

        if var1 + var2 == 0:
            t_scores[pos] = 0.0
        elif m1 > m2:
            t_scores[pos] = (m1 - m2) / sqrt(var1 + var2)
        else:
            t_scores[pos] = (m2 - m1) / sqrt(var1 + var2)

    cdef np.ndarray[DTYPE_INT_t] candidate_poss = np.argsort(
        t_scores).astype(DTYPE_INT)[::-1]

    cdef np.ndarray[DTYPE_INT_t] cpts = np.empty(num_cpts, dtype=DTYPE_INT)
    cpts[0] = candidate_poss[0] + running_stat_width
    blacklist_pos = set(range(
        candidate_poss[0] - min_base_obs + 1, candidate_poss[0] + min_base_obs))
    cdef DTYPE_INT_t cand_pos
    cdef DTYPE_INT_t cand_idx = 1
    cdef DTYPE_INT_t added_cpts = 1
    while added_cpts < num_cpts:
        cand_pos = candidate_poss[cand_idx]
        if cand_pos not in blacklist_pos:
            cpts[added_cpts] = cand_pos + running_stat_width
            added_cpts += 1
            blacklist_pos.update(range(
                cand_pos - min_base_obs + 1, cand_pos + min_base_obs))
        cand_idx += 1
        if cand_idx >= num_cands:
            raise NotImplementedError, 'Fewer changepoints found than requested.'

    return cpts

def c_calc_llh_ratio(
        np.ndarray[DTYPE_t] reg_means,
        np.ndarray[DTYPE_t] reg_ref_means, np.ndarray[DTYPE_t] reg_ref_vars,
        np.ndarray[DTYPE_t] reg_alt_means, np.ndarray[DTYPE_t] reg_alt_vars):
    cdef DTYPE_t ref_z_sum, ref_log_var_sum, alt_z_sum, alt_log_var_sum
    ref_z_sum, ref_log_var_sum, alt_z_sum, alt_log_var_sum = 0.0, 0.0 ,0.0 ,0.0
    cdef DTYPE_t ref_diff, alt_diff, log_lh_ratio
    cdef DTYPE_INT_t idx
    for idx in range(reg_means.shape[0]):
        ref_diff = reg_means[idx] - reg_ref_means[idx]
        ref_z_sum += (ref_diff * ref_diff) / reg_ref_vars[idx]
        ref_log_var_sum += log(reg_ref_vars[idx])

        alt_diff = reg_means[idx] - reg_alt_means[idx]
        alt_z_sum += (alt_diff * alt_diff) / reg_alt_vars[idx]
        alt_log_var_sum += log(reg_alt_vars[idx])

    log_lh_ratio = alt_z_sum + alt_log_var_sum - ref_z_sum - ref_log_var_sum

    return log_lh_ratio



from __future__ import division, unicode_literals, absolute_import

from builtins import int, range, dict, map, zip

import os
import io
import re
import sys
import random
import fnmatch

# Future warning from cython in h5py
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import h5py

import numpy as np

from glob import glob
from operator import itemgetter
from itertools import repeat, islice
from collections import defaultdict, namedtuple

if sys.version_info[0] > 2:
    unicode = str

# import tombo functions
from ._version import TOMBO_VERSION
from .c_helper import c_new_mean_stds, c_new_means, c_apply_outlier_thresh
from ._default_parameters import ROBUST_QUANTS, NUM_READS_FOR_SCALE

VERBOSE = False


################################
###### Global Namedtuples ######
################################

alignInfo = namedtuple(
    'alignInfo',
    ('ID', 'Subgroup', 'ClipStart', 'ClipEnd',
     'Insertions', 'Deletions', 'Matches', 'Mismatches'))

readData = namedtuple('readData', (
    'start', 'end', 'filtered', 'read_start_rel_to_raw',
    'strand', 'fn', 'corr_group', 'rna'))

intervalData = namedtuple('intervalData', (
    'reg_id', 'chrm', 'start', 'end', 'strand', 'reg_text', 'reads', 'seq'))
# set default values for strand, text, reads and seq
intervalData.__new__.__defaults__ = (None, '', None, None)

channelInfo = namedtuple(
    'channelInfo',
    ('offset', 'range', 'digitisation', 'number', 'sampling_rate'))

scaleValues = namedtuple(
    'scaleValues',
    ('shift', 'scale', 'lower_lim', 'upper_lim'))

genomeLoc = namedtuple(
    'genomeLoc', ('Start', 'Strand', 'Chrom'))

NORM_TYPES = ('none', 'pA', 'pA_raw', 'median', 'robust_median',
              'median_const_scale')

# single base conversion for motifs
SINGLE_LETTER_CODE = {
    'A':'A', 'C':'C', 'G':'G', 'T':'T', 'B':'[CGT]',
    'D':'[AGT]', 'H':'[ACT]', 'K':'[GT]', 'M':'[AC]',
    'N':'[ACGT]', 'R':'[AG]', 'S':'[CG]', 'V':'[ACG]',
    'W':'[AT]', 'Y':'[CT]'}
INVALID_BASES = re.compile('[^ACGT]')


######################################
###### Various Helper Functions ######
######################################

def _warning_message(message):
    sys.stderr.write(
        '*' * 20 + ' WARNING ' + '*' * 20 + '\n\t' +
        message + '\n')
    return

def _error_message_and_exit(message):
    sys.stderr.write(
        '*' * 20 + ' ERROR ' + '*' * 20 + '\n\t' +
        message + '\n')
    sys.exit()
    return

COMP_BASES = dict(zip(map(ord, 'ACGT'), map(ord, 'TGCA')))
def comp_seq(seq):
    """
    Complement DNA sequence
    """
    return seq.translate(COMP_BASES)
def rev_comp(seq):
    """
    Reverse complement DNA sequence
    """
    return seq.translate(COMP_BASES)[::-1]

U_TO_T = {ord('U'):ord('T')}
def rev_transcribe(seq):
    """
    Convert U bases to T
    """
    return seq.translate(U_TO_T)

def get_chrm_sizes(raw_read_coverage, raw_read_coverage2=None):
    """
    Get covered chromosome sizes from a set of reads
    """
    strand_chrm_sizes = defaultdict(list)
    for (chrm, strand), cs_read_cov in \
        raw_read_coverage.items():
        strand_chrm_sizes[chrm].append(max(
            r_data.end for r_data in cs_read_cov))
    if raw_read_coverage2 is not None:
        for (chrm, strand), cs_read_cov in \
            raw_read_coverage2.items():
            strand_chrm_sizes[chrm].append(max(
                r_data.end for r_data in cs_read_cov))

    return dict((chrm, max(strnd_sizes))
                for chrm, strnd_sizes in
                strand_chrm_sizes.items())

class TomboMotif(object):
    def _parse_motif(self, rev_comp_motif=False):
        """
        Parse a single letter code motif into a pattern for matching
        """
        conv_motif = ''.join(SINGLE_LETTER_CODE[letter]
                             for letter in self.raw_motif)
        if rev_comp_motif:
            # reverse complement and then flip any group brackets
            conv_motif = rev_comp(conv_motif).translate({
                ord('['):']', ord(']'):'['})
        return re.compile(conv_motif)

    def __init__(self, raw_motif, mod_pos=None):
        invalid_chars = re.findall(
            '[^' + ''.join(SINGLE_LETTER_CODE) + ']', raw_motif)
        if len(invalid_chars) > 0:
            _error_message_and_exit(
                'Invalid characters in motif: ' + ', '.join(invalid_chars))

        # basic motif parsing
        self.raw_motif = raw_motif
        self.motif_len = len(raw_motif)
        self.motif_pat = self._parse_motif()
        self.rev_comp_pat = self._parse_motif(True)

        self.is_palindrome = self.motif_pat == self.rev_comp_pat

        # parse modified position from motif if provided
        self.mod_pos = mod_pos
        if mod_pos is None:
            self.mod_base = None
        else:
            self.mod_base = raw_motif[mod_pos - 1]
            if INVALID_BASES.match(self.mod_base):
                _warning_message(
                    'Provided modified position is not a single base, which ' +
                    'is likely an error. Specified modified base is one of: ' +
                    ' '.join(SINGLE_LETTER_CODE[self.mod_base][1:-1]))


def invalid_seq(seq):
    return bool(INVALID_BASES.search(seq))


###########################
###### FASTA Parsing ######
###########################

class Fasta(object):
    """
    Fasta sequence format wrapper class.

    Will load faidx via pyfaidx package if installed, else the fasta will be
    loaded into memory for sequence extraction
    """
    def _load_in_mem(self):
        genome_index = {}
        curr_id = None
        curr_seq = []
        with io.open(self.fasta_fn) as fasta_fp:
            for line in fasta_fp:
                if line.startswith('>'):
                    if (curr_id is not None and
                        curr_seq is not []):
                        genome_index[curr_id] = ''.join(curr_seq)
                    curr_seq = []
                    curr_id = line.replace(">","").split()[0]
                else:
                    curr_seq.append(line.strip())
            # add last record
            if (curr_id is not None and
                curr_seq is not []):
                genome_index[curr_id] = ''.join(curr_seq)

        return genome_index

    def __init__(self, fasta_fn, dry_run=False, force_in_mem=False):
        self.fasta_fn = fasta_fn
        try:
            if force_in_mem: raise ImportError
            import pyfaidx
            self.has_pyfaidx = True
            self.index = pyfaidx.Faidx(fasta_fn)
        except:
            self.has_pyfaidx = False
            if not dry_run:
                self.index = self._load_in_mem()

    def get_seq(self, chrm, start=None, end=None):
        if self.has_pyfaidx:
            if not (start or end):
                return self.index.fetch(
                    chrm, 1, self.index.index[chrm].rlen).seq.upper()
            if start < 0 or end > self.index.index[chrm].rlen:
                raise NotImplementedError(
                    'Encountered invalid genome sequence request.')
            return self.index.fetch(chrm, start + 1, end).seq.upper()
        return self.index[chrm][start:end].upper()

    def iter_chrms(self):
        if self.has_pyfaidx:
            for chrm in self.index.index:
                yield unicode(chrm)
        else:
            for chrm in self.index:
                yield chrm


#############################################
###### Automatic Sample-Type Detection ######
#############################################

def is_read_rna(fast5_data):
    """
    Determine if a read is RNA or DNA
    """
    # check both experiment type and kit slots for "rna"
    exp_type, exp_kit = None, None
    try:
        exp_type = fast5_data['UniqueGlobalKey/context_tags'].attrs[
            'experiment_type']
        try:
            exp_type = exp_type.decode()
        except (AttributeError, TypeError):
            pass
        # remove the word internal since it contains rna.
        exp_type = exp_type.replace('internal', '')
    except:
        pass
    try:
        exp_kit = fast5_data['UniqueGlobalKey/context_tags'].attrs[
            'experiment_kit']
        try:
            exp_kit = exp_kit.decode()
        except (AttributeError, TypeError):
            pass
        # remove the word internal since it contains rna.
        exp_kit = exp_kit.replace('internal', '')
    except:
        pass

    if exp_type is None and exp_kit is None:
        rna = False
    else:
        rna = (
            (exp_type is not None and re.search('rna', exp_type) is not None) or
            (exp_kit is not None and re.search('rna', exp_kit) is not None))

    return rna

def is_rna(raw_read_coverage, n_reads=10):
    """
    Determine if a set of reads are RNA or DNA from a small sample
    """
    proc_reads = 0
    for cs_reads in raw_read_coverage.values():
        for r_data in cs_reads:
            if not r_data.rna:
                return False
            proc_reads += 1
            if proc_reads >= n_reads:
                break
        if proc_reads >= n_reads:
            break
    return True

def is_rna_from_files(fast5_fns, n_reads=10):
    """
    Determine if a set of files are RNA or DNA from a small sample
    """
    proc_reads = 0
    for fast5_fn in fast5_fns:
        try:
            with h5py.File(fast5_fn, 'r') as fast5_data:
                if not is_read_rna(fast5_data):
                    return False
            proc_reads += 1
        except:
            continue
        if proc_reads >= n_reads:
            break
    return True


#########################################
###### Index File/Filter Functions ######
#########################################

def get_index_fn(fast5s_dir, corr_grp):
    """
    Get the filename for the requested directory and corrected group
    """
    # if directory comes with trailing slash, remove for processing
    if fast5s_dir.endswith('/'):
        fast5s_dir = fast5s_dir[:-1]
    split_dir = os.path.split(fast5s_dir)
    return os.path.join(split_dir[0], "." + split_dir[1] +
                        "." + corr_grp + '.tombo.index')

def get_lock_fn(fast5s_dir):
    """
    Get filename for the lock file to indicate that this directory
    is currently being processed. This file should be saved to be deleted later.
    """
    # if directory comes with trailing slash, remove for processing
    if fast5s_dir.endswith('/'):
        fast5s_dir = fast5s_dir[:-1]
    split_dir = os.path.split(fast5s_dir)
    return os.path.join(split_dir[0], "." + split_dir[1] + '.tombo.lock')

def _is_lock_file(lock_fn):
    lock_bn = os.path.split(lock_fn)[1]
    if (len(lock_bn) == 0 or
        not lock_bn.startswith('.') or
        not lock_bn.endswith('.tombo.lock')):
        return False
    return True

def prep_index_data(
        fast5_fn, genome_loc, read_start_rel_to_raw, segs,
        corr_grp, subgroup, rna, obs_filter):
    """
    Prepare data for storage in the index file
    """
    if obs_filter is None:
        is_filtered = False
    else:
        base_lens = np.diff(segs)
        is_filtered = any(np.percentile(base_lens, pctl) > thresh
                          for pctl, thresh in obs_filter)
    mapped_end = genome_loc.Start + len(segs) - 1

    return ((genome_loc.Chrom, genome_loc.Strand), (
        fast5_fn, genome_loc.Start, mapped_end, read_start_rel_to_raw,
        corr_grp, subgroup, is_filtered, rna))

def write_index_file(all_index_data, index_fn, basedir):
    """
    Write index file
    """
    try:
        import cPickle as pickle
    except:
        import pickle
    index_data = defaultdict(list)
    for chrm_strand, (fn, start, end, rsrtr, c_grp, s_grp,
                      filtered, rna) in all_index_data:
        # clip the basedir off the FAST5 filename in case later functions are
        # called from another relative path
        from_base_fn = fn.replace(basedir, '')
        index_data[chrm_strand].append((
            from_base_fn, start, end, rsrtr, c_grp, s_grp, filtered, rna))

    with io.open(index_fn, 'wb') as index_fp:
        # note protocol 2 for py2/3 compatibility
        pickle.dump(dict(index_data), index_fp, protocol=2)

    return

def clear_filters(fast5s_dir, corr_grp):
    """
    Clear filters applied to this directories index files
    """
    fast5s_dir = (fast5s_dir if fast5s_dir.endswith('/') else
                  fast5s_dir + '/')
    index_fn = get_index_fn(fast5s_dir, corr_grp)
    try:
        import cPickle as pickle
    except:
        import pickle
    try:
        with io.open(index_fn, 'rb') as index_fp:
            index_data = pickle.load(index_fp)
    except IOError:
        _error_message_and_exit(
            'Filters can only be applied to runs ' +
            'with a Tombo index file. Re-run resquiggle without the ' +
            '--skip-index option to apply filters.')
    new_index_data = []
    for chrm_strand, cs_raw_data in index_data.items():
        new_index_data.extend([(chrm_strand, (
            from_base_fn, start, end, rsrtr, corr_grp, s_grp, False, rna))
                               for from_base_fn, start, end, rsrtr, c_grp,
                               s_grp, filtered, rna in cs_raw_data])

    write_index_file(new_index_data, index_fn, fast5s_dir)
    sys.stderr.write('All filters successfully cleared!\n')

    return

def parse_obs_filter(obs_filter):
    """
    Parse observations per base formatted filtering
    """
    if len(obs_filter) < 1:
        return None

    # parse obs_filter
    try:
        obs_filter = [list(map(int, pctl_nobs.split(':')))
                      for pctl_nobs in obs_filter]
    except:
        raise RuntimeError('Invalid format for observation filter')

    if any(pctl < 0 or pctl > 100 for pctl in map(itemgetter(0), obs_filter)):
       _error_message_and_exit('Invalid percentile value.')

    return obs_filter

def filter_reads(fast5s_dir, corr_grp, obs_filter):
    """
    Filter reads based on some observation per base threshold criteria
    """
    def read_is_stuck(fast5_fn, s_grp):
        try:
            with h5py.File(fast5_fn, 'r') as fast5_data:
                base_lens = fast5_data['/Analyses/' + corr_grp + '/' + s_grp +
                                       '/Events']['length']
                return any(np.percentile(base_lens, pctl) > thresh
                           for pctl, thresh in obs_filter)
        except:
            return True

    fast5s_dir = (fast5s_dir if fast5s_dir.endswith('/') else
                  fast5s_dir + '/')
    index_fn = get_index_fn(fast5s_dir, corr_grp)
    try:
        import cPickle as pickle
    except:
        import pickle
    try:
        with io.open(index_fn, 'rb') as index_fp:
            index_data = pickle.load(index_fp)
    except IOError:
        sys.stderr.write(
            '******** ERRROR *******\n\tFilters can only be applied to runs ' +
            'with a Tombo index file. Re-run resquiggle without the ' +
            '--skip-index option to apply filters.\n')
    filt_index_data = []
    num_reads, num_filt_reads = 0, 0
    for chrm_strand, cs_raw_data in index_data.items():
        cs_filt_reads = [
            (chrm_strand, (
                from_base_fn, start, end, rsrtr, corr_grp, s_grp,
                read_is_stuck(fast5s_dir + '/' + from_base_fn, s_grp), rna))
            for from_base_fn, start, end, rsrtr, c_grp,
            s_grp, filtered, rna in cs_raw_data if not filtered]
        num_reads += len(cs_raw_data)
        num_filt_reads += sum([1 for i_data in cs_filt_reads if i_data[1][6]])
        filt_index_data.extend(cs_filt_reads)

    sys.stderr.write(
        'Filtered ' + unicode(num_filt_reads) +
        ' reads due to observations per base filter from a ' +
        'total of ' + unicode(num_reads) + ' reads in ' + fast5s_dir + '.\n')

    write_index_file(filt_index_data, index_fn, fast5s_dir)

    return

def filter_reads_for_coverage(fast5s_dir, corr_grp, frac_to_filter):
    fast5s_dir = (fast5s_dir if fast5s_dir.endswith('/') else
                  fast5s_dir + '/')
    index_fn = get_index_fn(fast5s_dir, corr_grp)
    try:
        import cPickle as pickle
    except:
        import pickle
    try:
        with io.open(index_fn, 'rb') as index_fp:
            index_data = pickle.load(index_fp)
    except IOError:
        sys.stderr.write(
            '******** ERRROR *******\n\tFilters can only be applied to runs ' +
            'with a Tombo index file. Re-run resquiggle without the ' +
            '--skip-index option to apply filters.\n')
    unfilt_data = []
    unfilt_reads_cov = []
    prev_filt_data = []
    for chrm_strand, cs_raw_data in index_data.items():
        max_end = max(end for (_, _, end, _, _, _, _, _) in cs_raw_data)
        cs_coverage = np.zeros(max_end, dtype=np.int64)
        for (from_base_fn, start, end, rsrtr, c_grp,
             s_grp, filtered, rna) in cs_raw_data:
            if filtered: continue
            cs_coverage[start:end] += 1
        # now go through and compute coverage as well
        for (from_base_fn, start, end, rsrtr, c_grp,
             s_grp, filtered, rna) in cs_raw_data:
            if filtered:
                prev_filt_data.append((chrm_strand, (
                    from_base_fn, start, end, rsrtr,
                    c_grp, s_grp, filtered, rna)))
                continue
            # add approximate coverage from middle of read
            # faster than mean over the whole read
            unfilt_reads_cov.append(cs_coverage[start + ((end - start) // 2)])
            unfilt_data.append((chrm_strand, (
                from_base_fn, start, end, rsrtr, c_grp, s_grp, filtered, rna)))

    num_reads = len(unfilt_data)
    num_filt_reads = int(frac_to_filter * num_reads)
    sys.stderr.write(
        'Filtered ' + unicode(num_filt_reads) +
        ' reads due to observations per base filter from a ' +
        'total of ' + unicode(num_reads) + ' reads in ' + fast5s_dir + '.\n')

    # create probabilities array with coverage values normalized to sum to 1
    unfilt_reads_cov = np.array(unfilt_reads_cov, dtype=np.float)
    unfilt_reads_p = unfilt_reads_cov / unfilt_reads_cov.sum()
    filt_indices = np.random.choice(
        num_reads, size=num_filt_reads, replace=False, p=unfilt_reads_p)
    filt_index_data = [
        (chrm_strand, (from_base_fn, start, end, rsrtr, c_grp, s_grp, True, rna))
        for (chrm_strand, (from_base_fn, start, end, rsrtr, c_grp, s_grp, _, rna))
        in itemgetter(*filt_indices)(unfilt_data)]
    unfilt_index_data = list(itemgetter(*list(set(range(num_reads)).difference(
        filt_indices)))(unfilt_data))

    write_index_file(prev_filt_data + filt_index_data + unfilt_index_data,
                     index_fn, fast5s_dir)

    return


#####################################
###### FAST5 Parsing Functions ######
#####################################

def annotate_with_fastqs(fastq_fns, fast5s_read_ids, fastq_slot):
    if VERBOSE: sys.stderr.write('Annotating FAST5s with sequence from FASTQs.\n')
    for fastq_fn in fastq_fns:
        n_recs = 0
        been_warned_ids = False
        with io.open(fastq_fn) as fastq_fp:
            while True:
                fastq_rec = list(islice(fastq_fp, 4))
                # if record contains fewer than 4 lines this indicates the
                # EOF, so move to next file
                if len(fastq_rec) != 4: break
                # if sequence identifier line does not start with "@" or quality
                # score line does not start with a "+" the file may be
                # corrupted, so don't process any more records
                if (re.match('@', fastq_rec[0]) is None or
                    re.match('\+', fastq_rec[2]) is None):
                    _warning_message(
                        'Successfully parsed ' + unicode(n_recs) +
                        'FASTQ records from ' + fastq_fn + ' before ' +
                        'encountering an invalid record. The rest of ' +
                        'this file will not be processed.')
                    break

                # extract read_id from fastq (which should be the first text
                # after the "@" record delimiter up to the first white space or
                # underscore
                read_id = fastq_rec[0].split()[0].split('_')[0][1:]
                if read_id not in fast5s_read_ids:
                    if not been_warned_ids:
                        been_warned_ids = True
                        _warning_message(
                            'Some records from ' + fastq_fn + ' contain read ' +
                            'identifiers not found in any FAST5 files.')
                    continue

                with h5py.File(fast5s_read_ids[read_id]) as fast5_data:
                    bc_slot = fast5_data[fastq_slot]
                    bc_slot.create_dataset(
                        'Fastq', data=''.join(fastq_rec),
                        dtype=h5py.special_dtype(vlen=unicode))

    return

def reads_contain_basecalls(fast5_fns, bc_grp, num_reads):
    test_fns = random.sample(
        fast5_fns, num_reads) if len(fast5_fns) > num_reads else fast5_fns
    for fast5_fn in test_fns:
        try:
            with h5py.File(fast5_fn, 'r') as fast5_data:
                fast5_data['/Analyses/' + bc_grp]
        except:
            continue
        # if the basecall group is accessible for a single file return true
        return True

    # else if all tested reads did not contain the basecall group return False
    return False

def get_files_list(fast5s_dir):
    """
    Get all fast5 files recursively below this directory
    """
    all_fast5s = []
    # walk through directory structure searching for fast5 files
    for root, _, fns in os.walk(fast5s_dir):
        for fn in fnmatch.filter(fns, '*.fast5'):
            all_fast5s.append(os.path.join(root, fn))

    return all_fast5s

def clear_tombo_locks(lock_fns):
    """
    Clear all lock files
    """
    for lock_fn in lock_fns:
        # safegaurd against incorrect file passed to this function
        if not _is_lock_file(lock_fn):
            continue
        # try to remove lock files so that is some have been corrupted or
        # otherwise cause an error at least all those accessible will be cleared
        try:
            os.remove(lock_fn)
        except:
            pass

    return

def get_files_list_and_lock_dirs(fast5s_dir, ignore_locks):
    """
    Get all fast5 files recursively below this directory and add a Tombo lock
    file to indicate that this directory is currently being re-squiggled
    """
    all_fast5s = []
    lock_fns = []
    # walk through directory structure searching for fast5 files
    for root, _, fns in os.walk(fast5s_dir):
        lock_fn = get_lock_fn(root)
        if not ignore_locks and os.path.exists(lock_fn):
            clear_tombo_locks(lock_fns)
            _error_message_and_exit(
                'This set of reads is currently being processed by another ' +
                'resquiggle command. Multiple resquiggle commands cannot be ' +
                'run concurrently on a set of reads to avoid corrupting ' +
                'read files. If you are sure this set of reads is not being ' +
                'processed by another command (usually caused by previous ' +
                'unexpected exit) set the --ignore-read-locks flag.')
        lock_fns.append(lock_fn)
        try:
            # create empty file indicating this directory is locked
            open(lock_fn, 'w').close()
        except:
            clear_tombo_locks(lock_fns)
            _error_message_and_exit(
                'Could not write tombo lock file. Check that you have write ' +
                'permission within the specified [fast5_basedir].')

        for fn in fnmatch.filter(fns, '*.fast5'):
            all_fast5s.append(os.path.join(root, fn))

    return all_fast5s, lock_fns

def get_raw_read_slot(fast5_data):
    try:
        raw_read_slot = list(fast5_data['/Raw/Reads'].values())[0]
    except:
        raise NotImplementedError(
            'Raw data is not found in /Raw/Reads/Read_[read#]')

    return raw_read_slot

def prep_fast5_for_fastq(
        fast5_data, basecall_group, basecall_subgroup, overwrite):
    try:
        read_id = get_raw_read_slot(fast5_data).attrs['read_id']
        try:
            read_id = read_id.decode()
        except (AttributeError, TypeError):
            pass
    except:
        return None

    # if Analyses group doesn't exist yet, create it
    try:
        analyses_grp = fast5_data['/Analyses']
    except:
        analyses_grp = fast5_data.create_group('Analyses')

    # create Fastq slot, unless value exists and --overwrite is not set
    try:
        bc_grp = analyses_grp[basecall_group]
    except:
        bc_grp = analyses_grp.create_group(basecall_group)
        bc_subgrp = bc_grp.create_group(basecall_subgroup)
    else:
        if overwrite:
            del analyses_grp[basecall_group]
            bc_grp = analyses_grp.create_group(basecall_group)
            bc_subgrp = bc_grp.create_group(basecall_subgroup)
        else:
            raise NotImplementedError(
                basecall_group + ' exists and --overwrite is not set.')

    return read_id

def get_read_ids_and_prep_fastq_slot(
        fast5s_dir, basecall_group, basecall_subgroup, overwrite):
    """
    Extract read id from /Raw group and prep fastq slots for annotation with
    associated FASTQ files.
    """
    if VERBOSE: sys.stderr.write(
            'Preparing reads and extracting read identifiers.\n')
    been_warned_overwrite = False
    been_warned_unique = False
    fast5s_read_ids = {}
    # walk through directory structure searching for fast5 files
    for root, _, fns in os.walk(fast5s_dir):
        for fn in fnmatch.filter(fns, '*.fast5'):
            fast5_fn = os.path.join(root, fn)
            with h5py.File(fast5_fn) as fast5_data:
                try:
                    read_id = prep_fast5_for_fastq(
                        fast5_data, basecall_group, basecall_subgroup, overwrite)
                except NotImplementedError:
                    if VERBOSE and not been_warned_overwrite:
                        been_warned_overwrite = True
                        _warning_message(
                            'Basecalls exsit in ' + basecall_group + ' slot. ' +
                            'Set --overwrite option to overwrite these ' +
                            'basecalls in this slot.')
                    continue
            if read_id is None:
                continue
            if read_id in fast5s_read_ids:
                # Warn non-unique read_ids in directory
                if VERBOSE and not been_warned_unique:
                    been_warned_unique = True
                    _warning_message(
                        'Multiple FAST5 files contain the same read ' +
                        'identifiers. Ensure that FAST5 files are from ' +
                        'a single run.')
                continue

            fast5s_read_ids[read_id] = fast5_fn

    return fast5s_read_ids

def parse_fast5s_wo_index(fast5_basedirs, corr_grp, bc_subgrps, rna):
    """
    Parse re-squiggled reads data from a list of fast5 directories
    """
    def get_read_data(read_fn, fast5_data, bc_subgrp):
        corr_data = fast5_data['/'.join(('/Analyses', corr_grp, bc_subgrp))]

        align_data = dict(corr_data['Alignment'].attrs.items())
        read_start_rel_to_raw = corr_data['Events'].attrs[
            'read_start_rel_to_raw']
        chrm = align_data['mapped_chrom']
        strand = align_data['mapped_strand']
        try:
            chrm = chrm.decode()
            strand = strand.decode()
        except:
            pass

        return chrm, strand, readData(
            align_data['mapped_start'], align_data['mapped_end'],
            False, read_start_rel_to_raw, strand, read_fn,
            corr_grp + '/' + bc_subgrp, rna)


    files = [fn for fast5_basedir in fast5_basedirs
             for fn in get_files_list(fast5_basedir)]
    raw_read_coverage = defaultdict(list)
    for read_fn in files:
        try:
            with h5py.File(read_fn, 'r') as fast5_data:
                for bc_subgrp in bc_subgrps:
                    chrm, strand, r_data = get_read_data(
                        read_fn, fast5_data, bc_subgrp)
                    raw_read_coverage[(chrm, strand)].append(r_data)
        except:
            # ignore errors and process all reads that don't error
            continue

    return dict(raw_read_coverage)

def convert_index(index_data, fast5s_dir, corr_grp, new_corr_grp):
    """
    Convert an index and save under a new corrected group. Mostly for
    model_resquiggle
    """
    new_index_data = []
    for (chrm, strand), cs_raw_data in index_data.items():
        for (from_base_fn, start, end, rsrtr, c_grp, s_grp,
             filtered, rna) in cs_raw_data:
            if c_grp != corr_grp: continue
            new_index_data.append(((chrm, strand), (
                from_base_fn, start, end, rsrtr,
                new_corr_grp, s_grp, filtered, rna)))

    fast5s_dir = (fast5s_dir if fast5s_dir.endswith('/') else
                  fast5s_dir + '/')
    new_index_fn = get_index_fn(fast5s_dir, new_corr_grp)
    write_index_file(new_index_data, new_index_fn, fast5s_dir)

    return

def parse_fast5s_w_index(fast5s_dir, corr_grp, subgroups, new_corr_grp):
    """
    Use index file to parse information about a set of reads
    """
    fast5s_dir = (fast5s_dir if fast5s_dir.endswith('/') else
                  fast5s_dir + '/')
    index_fn = get_index_fn(fast5s_dir, corr_grp)
    try:
        import cPickle as pickle
    except:
        import pickle
    with io.open(index_fn, 'rb') as index_fp:
        index_data = pickle.load(index_fp)
    raw_read_coverage = {}
    for (chrm, strand), cs_raw_data in index_data.items():
        cs_data = [
            readData(start, end, filtered, rsrtr, strand,
                     os.path.join(fast5s_dir, from_base_fn),
                     corr_grp + '/' + s_grp, rna)
            for from_base_fn, start, end, rsrtr, c_grp,
            s_grp, filtered, rna in cs_raw_data
            if c_grp == corr_grp and s_grp in subgroups and not filtered]
        raw_read_coverage[(chrm, strand)] = cs_data
    if new_corr_grp is not None:
        # convert corrected group to new corrected group for
        # model re-squiggle
        convert_index(index_data, fast5s_dir, corr_grp, new_corr_grp)

    return raw_read_coverage

def merge_cov(w_index_covs, wo_index_cov):
    """
    Merge coverage from serveral parsed sets of data
    """
    all_covs = w_index_covs + [wo_index_cov,]
    raw_read_coverage = defaultdict(list)
    for chrm_strand in set([cs for d_cov in all_covs for cs in d_cov]):
        for dir_cov in all_covs:
            if chrm_strand not in dir_cov: continue
            raw_read_coverage[chrm_strand].extend(dir_cov[chrm_strand])

    return dict(raw_read_coverage)

def parse_fast5s(fast5_basedirs, corrected_group, basecall_subgroups,
                 new_corr_grp=None, rna=False):
    """
    Parse data from a list of re-squiggle fast5 directories
    """
    if VERBOSE: sys.stderr.write('Parsing tombo index file(s).\n')
    wo_index_dirs = []
    w_index_covs = []
    warn_index = False
    # determine if index exists for each directory and load appropriately
    for fast5s_dir in fast5_basedirs:
        fast5s_dir = (fast5s_dir if fast5s_dir.endswith('/') else
                      fast5s_dir + '/')
        if os.path.exists(get_index_fn(fast5s_dir, corrected_group)):
            try:
                w_index_covs.append(parse_fast5s_w_index(
                    fast5s_dir, corrected_group, basecall_subgroups,
                    new_corr_grp))
            except:
                raise
                _warning_message(
                    'Failed to parse tombo index file for ' +
                    fast5s_dir + ' directory.')
                wo_index_dirs.append(fast5s_dir)
        else:
            if not warn_index:
                _warning_message(
                    'Tombo index file does not exist for one or more ' +
                    'directories. If --skip-index was not set for ' +
                    're-squiggle command, ensure that the specified ' +
                    'directory is the same as for the re-squiggle command.\n')
                warn_index = True
            wo_index_dirs.append(fast5s_dir)
    wo_index_cov = parse_fast5s_wo_index(
        wo_index_dirs, corrected_group, basecall_subgroups, rna)
    raw_read_coverage = merge_cov(w_index_covs, wo_index_cov)

    return raw_read_coverage


###################################
###### Statistical Functions ######
###################################

# Some of these functions should likely be re-factored to tombo_stats

def parse_pore_model(pore_model_fn):
    """
    Parse pore model for pA normalization (Deprecated)
    """
    pore_model = {'mean':{}, 'inv_var':{}}
    with io.open(pore_model_fn) as fp:
        for line in fp:
            if line.startswith('#'): continue
            try:
                kmer, lev_mean, lev_stdev = line.split()[:3]
                lev_mean, lev_stdev = map(float, (lev_mean, lev_stdev))
            except ValueError:
                # header or other non-kmer field
                continue
            pore_model['mean'][kmer] = lev_mean
            pore_model['inv_var'][kmer] = 1 / (lev_stdev * lev_stdev)

    return pore_model

def calc_kmer_fitted_shift_scale(pore_model, events_means, events_kmers):
    """
    Compute fitted shift and scale parameters for pA normalization
    """
    r_model_means = np.array([pore_model['mean'][kmer]
                              for kmer in events_kmers])
    r_model_inv_vars = np.array([pore_model['inv_var'][kmer]
                                 for kmer in events_kmers])
    model_mean_var = r_model_means * r_model_inv_vars
    # prep kmer model coefficient matrix for the k-mers from this read
    model_mean_var_sum = model_mean_var.sum()
    coef_mat = np.array((
        (r_model_inv_vars.sum(), model_mean_var_sum),
        (model_mean_var_sum, (model_mean_var * r_model_means).sum())))

    # prep dependent values from this reads true events
    r_event_var = events_means * r_model_inv_vars
    r_event_var_mean = r_event_var * r_model_means
    dep_vect = np.array((r_event_var.sum(), r_event_var_mean.sum()))

    shift, scale = np.linalg.solve(coef_mat, dep_vect)

    return shift, scale

def get_valid_cpts(norm_signal, running_stat_width, num_events):
    """
    Get valid changepoints given largest differences in neighboring
    moving windows

    Note that this method is completely vectorized, but allows segments
    as small as 2 observations. This should be okay R9+, but is problematic
    for <=R7 and RNA
    """
    raw_cumsum = np.cumsum(np.concatenate([[0], norm_signal[:-1]]))
    # get difference between all neighboring running_stat_width regions
    running_diffs = np.abs(
        (2 * raw_cumsum[running_stat_width:-running_stat_width]) -
        raw_cumsum[:-2*running_stat_width] -
        raw_cumsum[2*running_stat_width:])
    not_peaks = np.logical_not(np.logical_and(
        running_diffs > np.concatenate([[0], running_diffs[:-1]]),
        running_diffs > np.concatenate([running_diffs[1:], [0]])))
    running_diffs[not_peaks] = 0
    valid_cpts = np.argsort(
        running_diffs)[::-1][:num_events].astype(np.int64) + running_stat_width

    return valid_cpts

def estimate_global_scale(fast5_fns, num_reads=NUM_READS_FOR_SCALE):
    sys.stderr.write('Estimating global scale parameter\n')
    np.random.shuffle(fast5_fns)
    read_mads = []
    for fast5_fn in fast5_fns:
        try:
            with h5py.File(fast5_fn, 'r') as fast5_data:
                all_sig = get_raw_read_slot(fast5_data)['Signal'].value
            shift = np.median(all_sig)
            read_mads.append(np.median(np.abs(all_sig - shift)))
        except:
            continue
        if len(read_mads) >= num_reads:
            break

    if len(read_mads) == 0:
        _error_message_and_exit(
            'No reads contain raw signal for ' +
            'global scale parameter estimation.')
    if len(read_mads) < num_reads:
        _warning_message(
            'Few reads contain raw signal for global scale parameter ' +
            'estimation. Results may not be optimal.')

    return np.mean(read_mads)

def normalize_raw_signal(
        all_raw_signal, read_start_rel_to_raw, read_obs_len,
        norm_type=None, channel_info=None, outlier_thresh=None,
        shift=None, scale=None, lower_lim=None, upper_lim=None,
        pore_model=None, event_means=None, event_kmers=None,
        const_scale=None):
    """
    Apply scaling and windsorizing parameters to normalize raw signal
    """
    if norm_type not in NORM_TYPES and (shift is None or scale is None):
        raise NotImplementedError(
            'Normalization type ' + norm_type + ' is not a valid ' +
            'option and shift or scale parameters were not provided.')

    raw_signal = np.array(all_raw_signal[
        read_start_rel_to_raw:
        read_start_rel_to_raw + read_obs_len])
    if shift is None or scale is None:
        if norm_type == 'none':
            shift, scale = 0, 1
        elif norm_type in ('pA_raw', 'pA'):
            # correct raw signal as described here:
            #     https://community.nanoporetech.com
            #           /posts/squiggle-plot-for-raw-data
            shift, scale = (
                -1 * channel_info.offset,
                channel_info.digitisation / channel_info.range)
            if norm_type == 'pA':
                # perform k-mer model fitted correction as in
                # nanocorr/nanopolish/ONT
                fit_shift, fit_scale = calc_kmer_fitted_shift_scale(
                    pore_model, event_means, event_kmers)
                # apply shift and scale values fitted from kmer
                # conditional model after raw DAC scaling
                shift = shift + (fit_shift * scale)
                scale = scale * fit_scale
        elif norm_type == 'median':
            shift = np.median(raw_signal)
            scale = np.median(np.abs(raw_signal - shift))
        elif norm_type == 'median_const_scale':
            assert const_scale is not None
            shift = np.median(raw_signal)
            scale = const_scale
        elif norm_type == 'robust_median':
            shift = np.mean(np.percentile(raw_signal, ROBUST_QUANTS))
            scale = np.median(np.abs(raw_signal - read_robust_med))

    raw_signal = (raw_signal - shift) / scale

    if outlier_thresh is not None or (
            lower_lim is not None and upper_lim is not None):
        if outlier_thresh is not None:
            read_med = np.median(raw_signal)
            read_mad = np.median(np.abs(raw_signal - read_med))
            lower_lim = read_med - (read_mad * outlier_thresh)
            upper_lim = read_med + (read_mad * outlier_thresh)
        raw_signal = c_apply_outlier_thresh(raw_signal, lower_lim, upper_lim)

    return raw_signal, scaleValues(shift, scale, lower_lim, upper_lim)


###########################################
###### Events Table Access Functions ######
###########################################

def get_multiple_slots_read_centric(r_data, slot_names, corr_grp=None):
    """
    Extract read-centric slot_names from this read's Events table
    """
    try:
        do_close = False
        if not isinstance(r_data, h5py.File):
            do_close = True
            corr_grp = r_data.corr_group
            r_data = h5py.File(r_data.fn, 'r')
        event_slot_name = '/'.join(('/Analyses', corr_grp, 'Events'))
        # note that it's more efficient to try to access the slot
        # and except the error that check if the slot exists first
        r_event_data = r_data[event_slot_name].value
        if do_close: r_data.close()
    except:
        # probably truncated file or events don't exist
        return [None,] * len(slot_names)

    return [r_event_data[slot_name] for slot_name in slot_names]

def get_single_slot_read_centric(r_data, slot_name, corr_grp=None):
    """
    Extract read-centric slot_name from this read's Events table
    """
    try:
        # if r_data is an open h5py object then don't open the filename
        do_close = False
        if not isinstance(r_data, h5py.File):
            do_close = True
            corr_grp = r_data.corr_group
            r_data = h5py.File(r_data.fn, 'r')
        # note that it's more efficient to try to access the slot
        # and except the error that check if the slot exists first
        r_slot_values = r_data['/'.join(('/Analyses', corr_grp, 'Events'))][
            slot_name]
        if do_close: r_data.close()
    except:
        # probably truncated file or events don't exist
        return None

    return r_slot_values

def get_single_slot_genome_centric(r_data, slot_name):
    """
    Extract genome-centric slot_name from this read's Events table
    """
    r_slot_values = get_single_slot_read_centric(r_data, slot_name)
    if r_slot_values is None:
        return None

    if r_data.strand == '-':
        r_slot_values = r_slot_values[::-1]

    return r_slot_values

def get_mean_slot_genome_centric(cs_reads, chrm_len, slot_name):
    """
    Get the mean over all reads at each covered genomic location for this
    slots value
    """
    base_sums = np.zeros(chrm_len)
    base_cov = np.zeros(chrm_len, dtype=np.int64)
    for r_data in cs_reads:
        # extract read means data so data across all chrms is not
        # in RAM at one time
        g_slot_values = get_single_slot_genome_centric(r_data, slot_name)
        if g_slot_values is None: continue

        base_sums[r_data.start:
                  r_data.start + len(g_slot_values)] += g_slot_values
        base_cov[r_data.start:r_data.start + len(g_slot_values)] += 1

    return base_sums / base_cov

def get_all_mean_slot_values(raw_read_coverage, chrm_sizes, slot_name):
    """
    Get the mean over all reads at each covered genomic location for this
    slots value over all covered chromosomes and strands
    """
    # ignore divide by zero errors that occur where there is no
    # coverage. Need to correct nan values after subtracting two sets of
    # coverage so leave as nan for now
    old_err_settings = np.seterr(all='ignore')
    # take the mean over all signal overlapping each base
    all_mean_values = {}
    for chrm, strand in [(c, s) for c in chrm_sizes for s in ('+', '-')]:
        if (chrm, strand) in raw_read_coverage:
            cs_mean_values = get_mean_slot_genome_centric(
                raw_read_coverage[(chrm, strand)], chrm_sizes[chrm], slot_name)
        else:
            cs_mean_values = np.empty(chrm_sizes[chrm])
            cs_mean_values[:] = np.nan
        all_mean_values[(chrm, strand)] = cs_mean_values
    _ = np.seterr(**old_err_settings)

    return all_mean_values

def get_all_mean_levels(raw_read_coverage, chrm_sizes):
    """
    Helper function to compute genome location mean levels
    """
    return get_all_mean_slot_values(raw_read_coverage, chrm_sizes, 'norm_mean')

def get_all_mean_stdev(raw_read_coverage, chrm_sizes):
    """
    Helper function to compute genome location mean levels
    """
    return get_all_mean_slot_values(raw_read_coverage, chrm_sizes, 'norm_stdev')

def get_all_mean_lengths(raw_read_coverage, chrm_sizes):
    """
    Helper function to compute genome location mean lengths
    """
    return get_all_mean_slot_values(raw_read_coverage, chrm_sizes, 'length')


###########################################
###### Special Data Access Functions ######
###########################################

def get_channel_info(fast5_data):
    """
    Get channel information for a read
    """
    try:
        fast5_info = fast5_data['UniqueGlobalKey/channel_id'].attrs
    except:
        raise NotImplementedError("No channel_id group in HDF5 file. " +
                                  "Probably mux scan HDF5 file.")

    channel_info = channelInfo(
        fast5_info['offset'], fast5_info['range'],
        fast5_info['digitisation'], fast5_info['channel_number'],
        fast5_info['sampling_rate'].astype(np.int64))

    return channel_info

def get_raw_signal(r_data, int_start, int_end):
    """
    Extract raw signal from where this read overlaps a particular genomic region
    """
    with h5py.File(r_data.fn, 'r') as fast5_data:
        # retrieve shift and scale computed in correction script
        corr_subgrp = fast5_data['/Analyses/' + r_data.corr_group]
        event_starts = corr_subgrp['Events']['start']
        events_end = event_starts[-1] + corr_subgrp['Events']['length'][-1]
        segs = np.concatenate([event_starts, [events_end,]])

        shift = corr_subgrp.attrs['shift']
        scale = corr_subgrp.attrs['scale']
        lower_lim = corr_subgrp.attrs['lower_lim']
        upper_lim = corr_subgrp.attrs['upper_lim']
        all_sig = get_raw_read_slot(fast5_data)['Signal'].value

    rsrtr = r_data.read_start_rel_to_raw
    if r_data.rna:
        # reverse raw signal for RNA
        all_sig = all_sig[::-1]

    if r_data.strand == "-":
        segs = (segs[::-1] * -1) + segs[-1]

    if int_start < r_data.start:
        # handle reads that start in the middle of the interval
        start_offset = r_data.start - int_start
        overlap_seg_data = segs[:int_end - r_data.start + 1]
    else:
        start_offset = 0
        skipped_bases = int_start - r_data.start
        overlap_seg_data = segs[
            skipped_bases:int_end - r_data.start + 1]

    num_reg_obs = overlap_seg_data[-1] - overlap_seg_data[0]
    if r_data.strand == "+":
        reg_start_rel_raw = rsrtr + overlap_seg_data[0]
        r_sig, _ = normalize_raw_signal(
            all_sig, reg_start_rel_raw, num_reg_obs, shift=shift,
            scale=scale, lower_lim=lower_lim, upper_lim=upper_lim)
    else:
        reg_start_rel_raw = rsrtr + segs[-1] - overlap_seg_data[-1]
        r_sig, _ = normalize_raw_signal(
            all_sig, reg_start_rel_raw, num_reg_obs, shift=shift,
            scale=scale, lower_lim=lower_lim, upper_lim=upper_lim)
        r_sig = r_sig[::-1]

    return r_sig, overlap_seg_data, start_offset

def parse_read_correction_data(r_data):
    """
    Parse correction data from an event resquiggled read
    """
    try:
        with h5py.File(r_data.fn, 'r') as fast5_data:
            corr_grp = fast5_data['/Analyses/' + r_data.corr_group]
            events_grp = corr_grp['Events']
            event_starts = events_grp['start']
            events_end = event_starts[-1] + events_grp['length'][-1]
            new_segs = np.concatenate([event_starts, [events_end,]])

            raw_grp = get_raw_read_slot(fast5_data)
            read_id = raw_grp.attrs['read_id']
            try:
                read_id = read_id.decode()
            except (AttributeError, TypeError):
                pass
            signal_data = raw_grp['Signal'].value

            raw_offset = events_grp.attrs['read_start_rel_to_raw']
            shift, scale, lower_lim, upper_lim = [
                corr_grp.attrs[attr_name] for attr_name in (
                    'shift', 'scale', 'lower_lim', 'upper_lim')]

            old_segs = corr_grp['Alignment/read_segments'].value
            old_align_vals = list(map(
                lambda x: x.decode(),
                corr_grp['Alignment/read_alignment'].value))
            new_align_vals = list(map(
                lambda x: x.decode(),
                corr_grp['Alignment/genome_alignment'].value))
    except:
        return None

    if r_data.rna:
        signal_data = signal_data[::-1]

    return (read_id, signal_data, raw_offset, shift, scale, lower_lim,
            upper_lim, old_segs, old_align_vals, new_align_vals,
            events_end, new_segs)

def get_all_read_data(r_data):
    """
    Extract most relevant read data from this read
    """
    try:
        with h5py.File(r_data.fn, 'r') as fast5_data:
            # note that it's more efficient to try to access the slot
            # and except the error that check if the slot exists first
            corr_subgrp = fast5_data['/Analyses/' + r_data.corr_group]
            algn_subgrp = dict(corr_subgrp['Alignment'].attrs.items())
            event_data = corr_subgrp['Events'].value
            r_attrs = dict(corr_subgrp.attrs.items())
            all_sig = get_raw_read_slot(fast5_data)['Signal'].value
    except:
        # probably truncated file or Events slot doesn't exist
        return None

    if r_data.rna:
        all_sig = all_sig[::-1]
    r_means = event_data['norm_mean']
    r_seq = b''.join(event_data['base']).decode()

    events_end = event_data[-1]['start'] + event_data[-1]['length']
    segs = np.concatenate([event_data['start'], [events_end,]]).astype(np.int64)
    r_sig, scale_values = normalize_raw_signal(
        all_sig, r_data.read_start_rel_to_raw, segs[-1] - segs[0],
        shift=r_attrs['shift'], scale=r_attrs['scale'],
        lower_lim=r_attrs['lower_lim'], upper_lim=r_attrs['upper_lim'])

    return (r_means, r_seq, r_sig, segs, scale_values,
            r_attrs['norm_type'], r_attrs['outlier_threshold'],
            genomeLoc(algn_subgrp['mapped_start'], algn_subgrp['mapped_strand'],
                      algn_subgrp['mapped_chrom']))

def get_coverage(raw_read_coverage):
    """
    Get genome coverage for a set of reads
    """
    if VERBOSE: sys.stderr.write('Calculating read coverage.\n')
    read_coverage = {}
    for (chrm, strand), reads_data in raw_read_coverage.items():
        if len(reads_data) == 0: continue
        max_end = max(r_data.end for r_data in reads_data)
        chrm_coverage = np.zeros(max_end, dtype=np.int64)
        for r_data in reads_data:
            chrm_coverage[r_data.start:r_data.end] += 1
        read_coverage[(chrm, strand)] = chrm_coverage

    return read_coverage

def get_reads_events(cs_reads):
    """
    Extract read base levels split by genomic position
    """
    # note that this function assumes that all reads come from the same
    # chromosome and strand
    cs_base_means = []
    cs_read_start_ends = []
    for r_data in cs_reads:
        # extract read means data so data across all chrms is not
        # in RAM at one time
        read_means = get_single_slot_genome_centric(r_data, 'norm_mean')
        if read_means is None: continue
        assert read_means.shape[0] == r_data.end - r_data.start, (
            'Read found with mismatching mapping location and ' +
            'signal information.')
        cs_base_means.append(read_means)
        cs_read_start_ends.append((r_data.start, r_data.end))

    if len(cs_base_means) == 0: return None

    chrm_signal = np.concatenate(cs_base_means)
    chrm_pos = np.concatenate([np.arange(start, end)
                               for start, end in cs_read_start_ends])
    # get order of all bases from position array
    as_chrm_pos = np.argsort(chrm_pos)
    # then sort the signal array by genomic position and
    # split into event means by base
    split_poss = np.where(
        np.concatenate([[0,], np.diff(
            chrm_pos[as_chrm_pos])]) > 0)[0]
    cs_base_events = dict(zip(
        np.unique(chrm_pos[as_chrm_pos]),
        np.split(chrm_signal[as_chrm_pos], split_poss)))

    return cs_base_events

def update_seq(r_data, reg_base_data, int_start, int_end):
    """
    Update the sequence for the region based on this read
    """
    r_seq = b''.join(get_single_slot_read_centric(r_data, 'base')).decode()
    if r_seq is None:
        # probably a corrupt file so return that the region is only
        # up to the start of this read so the next valid read will be added
        # to the sequence
        return reg_base_data, max(0, r_data.start - int_start)
    if r_data.strand == '-':
        r_seq = rev_comp(r_seq)

    # if read starts before the interval
    if r_data.start <= int_start:
        r_end_overlap = r_data.end - int_start
        # if read covers the whole interval
        if r_data.end > int_end:
            r_end_clip = r_data.end - int_end
            reg_base_data = r_seq[-r_end_overlap:-r_end_clip]
            return reg_base_data, len(reg_base_data)
        # end of read overlaps beginning of interval
        reg_base_data[:r_end_overlap] = r_seq[-r_end_overlap:]
        return reg_base_data, r_end_overlap
    # read doesn't cover the beginning of region
    if r_data.end > int_end:
        # beginning of read covers to the end of the region
        r_begin_overlap = int_end - r_data.start
        reg_base_data[-r_begin_overlap:] = r_seq[:r_begin_overlap]
        return reg_base_data, len(reg_base_data)
    # first read is completely contained in the interval
    r_len = r_data.end - r_data.start
    r_int_start = r_data.start - int_start
    reg_base_data[r_int_start:r_int_start + r_len] = r_seq
    return reg_base_data, r_int_start + r_len

def get_seq_from_reads(int_start, int_end, reg_reads):
    """
    Extract the forward strand genomic sequence for an interval from
    a set of reads
    """
    # handle case where no read overlaps whole region
    # let each read contibute its sequence and fill the rest
    # with dashes
    reg_base_data = ['-'] * (int_end - int_start)
    if len(reg_reads) == 0:
        return ''.join(reg_base_data)
    # get region sequence by moving through reads that
    # cover the region, but only extract seqeunce from the
    # (close to) minimal reads
    s_reg_reads = sorted(reg_reads, key=lambda r: (r.start, r.end))
    # begin filling sequence with first (by start pos) read
    reg_base_data, curr_cov_pos = update_seq(
        s_reg_reads.pop(0), reg_base_data, int_start, int_end)
    # if there was only one read return now
    if len(s_reg_reads) == 0 or curr_cov_pos >= int_end - int_start:
        return ''.join(reg_base_data)

    # get next read (by start pos)
    curr_read = s_reg_reads.pop(0)
    for next_read in s_reg_reads:
        # once the next read start passes the region covered thus far
        # add the sequence for the saved curr_read to the reg sequence
        if next_read.start >= curr_cov_pos:
            # add read with curr longest end position to the region seq
            reg_base_data, curr_cov_pos = update_seq(
                curr_read, reg_base_data, int_start, int_end)
            curr_read = next_read
            # if the whole interval is covered return the sequence
            if curr_cov_pos >= int_end - int_start:
                return ''.join(reg_base_data)
            continue
        if next_read.end > curr_read.end:
            curr_read = next_read

    reg_base_data, _ = update_seq(
        curr_read, reg_base_data, int_start, int_end)

    return ''.join(reg_base_data)

def add_reg_seq(all_reg_data):
    """
    Add the region sequence to the region data by extraction from a minimal
    set of reads
    """
    all_reg_base_data = []
    for reg_data in all_reg_data:
        # add text to each regions data
        all_reg_base_data.append(reg_data._replace(seq=get_seq_from_reads(
            reg_data.start, reg_data.end, reg_data.reads)))

    return all_reg_base_data

def get_region_reads(
        plot_intervals, raw_read_coverage, filter_no_cov=True, add_seq=True):
    """
    Get all reads overlapping a set of intervals
    """
    def get_c_s_data(chrm, strand, start, end):
        # get all reads intersecting the interval
        if (chrm, strand) in raw_read_coverage:
            r_data = raw_read_coverage[(chrm, strand)][0]
            return [
                r_data for r_data in raw_read_coverage[(chrm, strand)]
                if not (r_data.start >= end or r_data.end <= start)]
        return []


    all_reg_data = []
    for int_i in plot_intervals:
        # get all reads that overlap this interval
        # note that this includes partial overlaps as these contribute
        # to coverage and other statistics so can't really restrict to
        # full coverage as previous versions of code did
        if int_i.strand is None:
            # if strand is None, get data from both strands
            all_reg_data.append(int_i._replace(
                reads=get_c_s_data(int_i.chrm, '+', int_i.start, int_i.end) +
                get_c_s_data(int_i.chrm, '-', int_i.start, int_i.end)))
        else:
            all_reg_data.append(int_i._replace(
                reads=get_c_s_data(int_i.chrm, int_i.strand,
                                   int_i.start, int_i.end)))

    if add_seq:
        all_reg_data = add_reg_seq(all_reg_data)
    if not filter_no_cov:
        return all_reg_data

    # filter out no coverage regions
    all_reg_data = [
        reg_data for reg_data in all_reg_data if len(reg_data.reads) > 0]

    no_cov_regions = [
        (len(reg_data.reads) == 0, unicode(reg_data.chrm) + ':' +
         unicode(reg_data.start))
        for reg_data in all_reg_data]
    if any(no_cov[0] for no_cov in no_cov_regions):
        _warning_message(
            'No coverage in regions: ' + '; '.join([
                reg for no_cov, reg in no_cov_regions if no_cov]))

    return all_reg_data

def get_region_sequences(
        plot_intervals, raw_read_coverage1, raw_read_coverage2=None):
    """
    Get the sequence for a set of intervals from a set of reads
    """
    all_reg_data = get_region_reads(
        plot_intervals, raw_read_coverage1, filter_no_cov=False, add_seq=False)
    if raw_read_coverage2 is not None:
        all_reg_data2 = get_region_reads(
            plot_intervals, raw_read_coverage2, filter_no_cov=False,
            add_seq=False)
        all_reg_data = [r1._replace(reads=r1.reads + r2.reads)
                        for r1, r2 in zip(all_reg_data, all_reg_data2)]
    all_reg_data = add_reg_seq(all_reg_data)

    return all_reg_data


###################################
###### FAST5 Write Functions ######
###################################

def prep_fast5(fast5_fn, corr_grp, overwrite, in_place,
               bc_grp=None, return_fp=False):
    """
    Prepare a read for re-squiggle processing (This deletes old re-squiggle
    info for this read)
    """
    def try_close_prep_err(fast5_data, err_str):
        try:
            fast5_data.close()
        except:
            pass
        return (err_str, fast5_fn)

    # several checks to prepare the FAST5 file for correction before
    # processing to save compute
    if not in_place:
        return ('Not currently implementing new hdf5 file writing',
                fast5_fn)
    # check that the file is writeable before trying to correct
    if not os.access(fast5_fn, os.W_OK):
        return ('FAST5 file is not writable', fast5_fn)

    try:
        # create group to store data
        fast5_data = h5py.File(fast5_fn, 'r+')
        try:
            analyses_grp = fast5_data['/Analyses']
        except:
            return try_close_prep_err(
                fast5_data, 'Analysis group not found at root of FAST5')
        try:
            # check that the requested basecalls group exsists
            if bc_grp is not None:
                analyses_grp[bc_grp]
        except:
            return try_close_prep_err(
                fast5_data, 'Basecall group not found at [--basecall-group]')

        try:
            corr_grp_ptr = analyses_grp[corr_grp]
            if not overwrite:
                return try_close_prep_err(
                    fast5_data, "Tombo data exsists in [--corrected-group] " +
                    "and [--overwrite] is not set")
            del analyses_grp[corr_grp]
        except:
            # if the corr_grp isn't there we will write it now, but
            # it's more efficient to try than to check if the slot is there
            pass

        corr_grp = analyses_grp.create_group(corr_grp)
        corr_grp.attrs['tombo_version'] = TOMBO_VERSION
        corr_grp.attrs['basecall_group'] = bc_grp
    except:
        return (
            'Error opening or writing to fast5 file', fast5_fn)

    if return_fp:
        return fast5_data

    try:
        fast5_data.close()
    except:
        return 'Error closing fast5 file', fast5_fn

    return

def write_error_status(
        filename, corrected_group, basecall_subgroup, error_text):
    """
    Write error message for a read into the FAST5 file
    """
    with h5py.File(filename, 'r+') as fast5_data:
        analysis_grp = fast5_data['/Analyses']
        corr_grp = analysis_grp[corrected_group]
        if basecall_subgroup is not None:
            # add subgroup matching subgroup from original basecalls
            corr_subgrp = corr_grp.create_group(basecall_subgroup)
            corr_subgrp.attrs['status'] = error_text
        else:
            corr_grp.attrs['status'] = error_text

    return

def write_new_fast5_group(
        fast5_data, genome_location, read_start_rel_to_raw,
        new_segs, align_seq, norm_signal, scale_values, corrected_group,
        basecall_subgroup, norm_type, outlier_thresh, compute_sd,
        alignVals=None, align_info=None, old_segs=None, rna=False):
    """
    Write new fast5 group with re-squiggle data
    """
    try:
        # compute event data before accessing fast5 file
        if compute_sd:
            norm_means, norm_stds = c_new_mean_stds(norm_signal, new_segs)
        else:
            norm_means = c_new_means(norm_signal, new_segs)
            norm_stds = repeat(np.NAN)

        # had to shift to names formats numpy array specification due to
        # python2 numpy unicode issues. See discussion here:
        # https://github.com/numpy/numpy/issues/2407
        event_data = np.array(
            list(zip(norm_means, norm_stds,
                     new_segs[:-1], np.diff(new_segs), list(align_seq))),
            dtype=[(str('norm_mean'), 'f8'), (str('norm_stdev'), 'f8'),
                   (str('start'), 'u4'), (str('length'), 'u4'),
                   (str('base'), 'S1')])

        if alignVals is not None:
            r_align_vals, g_align_vals = zip(*alignVals)
            np_read_align = np.chararray(len(alignVals))
            np_read_align[:] = r_align_vals
            np_genome_align = np.chararray(len(alignVals))
            np_genome_align[:] = g_align_vals
    except:
        raise NotImplementedError('Error computing new events')

    do_close = False
    if not isinstance(fast5_data, h5py.File):
        try:
            fast5_data = h5py.File(fast5_data, 'r+')
            do_close = True
        except:
            raise NotImplementedError(
                'Error opening file for new group writing. This should ' +
                'have been caught during the alignment phase. Check that ' +
                'there are no other tombo processes or processes ' +
                'accessing these HDF5 files running simultaneously.')

    try:
        analysis_grp = fast5_data['/Analyses']
        corr_grp = analysis_grp[corrected_group]
        # add subgroup matching subgroup from original basecalls
        corr_subgrp = corr_grp.create_group(basecall_subgroup)
        corr_subgrp.attrs['status'] = 'success'
        corr_subgrp.attrs['rna'] = rna
        corr_subgrp.attrs['shift'] = scale_values.shift
        corr_subgrp.attrs['scale'] = scale_values.scale
        corr_subgrp.attrs['lower_lim'] = scale_values.lower_lim
        corr_subgrp.attrs['upper_lim'] = scale_values.upper_lim
        corr_subgrp.attrs['norm_type'] = norm_type
        corr_subgrp.attrs['outlier_threshold'] = outlier_thresh

        # store alignment statistics
        corr_alignment = corr_subgrp.create_group('Alignment')
        corr_alignment.attrs['mapped_start'] = genome_location.Start
        corr_alignment.attrs['mapped_end'] \
            = genome_location.Start + len(new_segs) - 1
        corr_alignment.attrs['mapped_strand'] = genome_location.Strand
        corr_alignment.attrs['mapped_chrom'] = genome_location.Chrom

        if align_info is not None:
            corr_alignment.attrs['clipped_bases_start'] = align_info.ClipStart
            corr_alignment.attrs['clipped_bases_end'] = align_info.ClipEnd
            corr_alignment.attrs['num_insertions'] = align_info.Insertions
            corr_alignment.attrs['num_deletions'] = align_info.Deletions
            corr_alignment.attrs['num_matches'] = align_info.Matches
            corr_alignment.attrs['num_mismatches'] = align_info.Mismatches

        if alignVals is not None:
            corr_alignment.create_dataset(
                'read_alignment', data=np_read_align, compression="gzip")
            corr_alignment.create_dataset(
                'genome_alignment', data=np_genome_align, compression="gzip")
        if old_segs is not None:
            # store old segmentation in order to plot "correction process"
            corr_alignment.create_dataset(
                'read_segments', data=old_segs, compression="gzip")

        # Add Events to data frame with event means, SDs and lengths
        corr_events = corr_subgrp.create_dataset(
            'Events', data=event_data, compression="gzip")
        corr_events.attrs[
            'read_start_rel_to_raw'] = read_start_rel_to_raw
    except:
        raise NotImplementedError(
            'Error writing resquiggle information back into fast5 file.')

    if do_close:
        try:
            fast5_data.close()
        except:
            raise NotImplementedError(
                'Error closing fast5 file after writing resquiggle information.')

    return


###################################
###### Filter Main Functions ######
###################################

def clear_filters_main(args):
    global VERBOSE
    VERBOSE = not args.quiet

    for fast5s_dir in args.fast5_basedirs:
        clear_filters(fast5s_dir, args.corrected_group)

    return

def filter_stuck_main(args):
    global VERBOSE
    VERBOSE = not args.quiet

    obs_filter = parse_obs_filter(args.obs_per_base_filter)
    for fast5s_dir in args.fast5_basedirs:
        filter_reads(fast5s_dir, args.corrected_group, obs_filter)

    return

def filter_coverage_main(args):
    global VERBOSE
    VERBOSE = not args.quiet

    if not 0 < args.percent_to_filter < 100:
        _error_message_and_exit(
            '--percent-to-filter must be between 0 and 100.')

    for fast5s_dir in args.fast5_basedirs:
        filter_reads_for_coverage(
            fast5s_dir, args.corrected_group, args.percent_to_filter / 100.0)

    return


##################################
###### Annotate FAST5s Main ######
##################################

def annotate_reads_with_fastq_main(args):
    global VERBOSE
    VERBOSE = not args.quiet

    fast5s_basedir = (
        args.fast5_basedir if args.fast5_basedir.endswith('/') else
        args.fast5_basedir + '/')
    fast5s_read_ids = get_read_ids_and_prep_fastq_slot(
        fast5s_basedir, args.basecall_group, args.basecall_subgroup,
        args.overwrite)
    fastq_slot = '/'.join(('/Analyses', args.basecall_group,
                           args.basecall_subgroup))
    annotate_with_fastqs(args.fastq_filenames, fast5s_read_ids, fastq_slot)

    return


if __name__ == '__main__':
    raise NotImplementedError(
        'This is a module. See commands with `tombo -h`')
